/* -----------------------------------------

   Test of HELL

   ----------------------------------------- */
#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/time.h> 

#include "hell.hh"
#include "h_coefficients.hh"
#include "Pqg.hh"
#include "CoefficientFunctions.hh"
#include "CoefficientFunctionsNew.hh"
#include "DISinterpolator.hh"

using namespace std;


namespace HELL {
  extern double integrandGSLx(double x, void *pars);
  extern dcomplex newDeltaFRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, process proc, int ordersub, void *pars);
};
struct s_prdtv { HELL::process p; double d1, d2, d3; HELL::xTableU *tU; void *par; };


int main (int argc, char* argv[]) {

  struct timeval t0, t1;
  gettimeofday(&t0,NULL);

  HELL::SetLLpMode(false);
  int nf = 4;
  HELL::HELLnf sxD(nf, HELL::NLL, "./data/");
  //sxD.SetRCmode(1);
  sxD.SetLargexDamping();
  sxD.SetMomentumConservation();
  sxD.SetEvaluationMethod(1);
  sxD.SetFactorizationScale(1);
  int Npoints = 9, Nas = 3;
  double as = 0.2;
  //double as = 0.115;
  double x = 1e-6;
  double mQ = 1.4/4.5;// *1e-5;

  /*
  mQ=2;
  HELL::Interpolator *Interp = new HELL::Interpolator(mQ*mQ);
  HELL::hCoeff hh2(HELL::MC2cc,Interp);
  cout << hh2.hk(0) << endl;
  cout << hh2.hk(1) << endl;
  HELL::hCoeff hhL(HELL::MCLcc,Interp);
  cout << hhL.hk(0) << endl;
  cout << hhL.hk(1) << endl;
  HELL::hCoeff hh3(HELL::MC3cc,Interp);
  cout << hh3.hk(0) << endl;
  cout << hh3.hk(1) << endl;
  exit(0);
  */

  /*
  dcomplex Delta_gammaLO_BatNLO = 1.;
  for(int i=0; i<20; i++) {
    Delta_gammaLO_BatNLO -= 0.1;
    HELL::DeltaCtest2Res(Delta_gammaLO_BatNLO, Delta_gammaLO_BatNLO/as, 0.6, as, nf);
    HELL::DeltaCtest1Res(Delta_gammaLO_BatNLO, Delta_gammaLO_BatNLO/as, 0.6, as, nf);
    HELL::oldDeltaCLRes(Delta_gammaLO_BatNLO, Delta_gammaLO_BatNLO/as, 0.6, as, nf);
    //cout << HELL::DeltaCLRes(Delta_gammaLO_BatNLO, Delta_gammaLO_BatNLO/as, 0.6, as, nf) << endl;
  }
  exit(0);
  */

  /*
  //sxD.SetEvaluationMethod(1);
  int _damping=2, _dampingsqrt=4;
  HELL::xTableU *tU = new HELL::xTableU("data/x-space-aux/xtableU_nlo_nf4_alphas0200.table");
  //HELL::xTableU *tU = new HELL::xTableU("data/x-space-aux/xtableUvar_nlo_nf4_alphas0200.table");
  cout << x*sxD.deltaCLg (as, x) << endl;
  cout << DeltaCLRes(x, as, nf, tU)*pow(1-x,_damping) * pow(1-sqrt(x),_dampingsqrt) << endl;
  //
  cout << x*sxD.deltaC2g (as, x) << endl;
  cout << DeltaC2Res(x, as, nf, tU)*pow(1-x,_damping) * pow(1-sqrt(x),_dampingsqrt) << endl;
  //
  HELL::Interpolator *Interp = new HELL::Interpolator(4*mQ*mQ); // if it's a new value of m/Q, then recompute it
  cout << x*sxD.deltaMCLg (as, x, mQ) << endl;
  cout << DeltaMCLRes(x, as, nf, tU, Interp)*pow(1-x,_damping) * pow(1-sqrt(x),_dampingsqrt) << endl;
  //
  //HiggsSmallX * ggHsx = new HiggsSmallX();
  ofstream ofile("output/testU.dat");
  int Np=500;
  for(int i=0; i<Np; i++) {
    x = 1e-9 * exp(i/(Np-1.)*log(0.999/1e-9));
    //s_prdtv intpars = { HELL::CL, 2e-8, as, HELL::beta0(nf), tU, NULL };
    s_prdtv intpars = { HELL::C2, 2.04878e-08, as, HELL::beta0(nf), tU, NULL };
    double Higgs_integrand = 0;
    //HELL::integrandCUBA(x,&intpars)
    ofile << setw(15) << x
	  << setw(15) << x*sxD.deltaCLg (as, x)
	  << setw(15) << DeltaCLRes(x, as, nf, tU)*pow(1-x,_damping) * pow(1-sqrt(x),_dampingsqrt)
	  << setw(15) << x*sxD.deltaC2g (as, x)
	  << setw(15) << DeltaC2Res(x, as, nf, tU)*pow(1-x,_damping) * pow(1-sqrt(x),_dampingsqrt)
	  << setw(15) << HELL::integrandGSLx(x,&intpars)
	  << setw(15) << DeltaUqgRes(x, as, nf, tU)*pow(1-x,_damping) * pow(1-sqrt(x),_dampingsqrt)
	  << setw(15) << Higgs_integrand
	  << endl;
  }
  exit(0);
  //*/

  /*
  //x = 1e-9;
  as=0.035;
  double rH = x*sxD.deltaHggAux(as, x, 125, 173, -4.7);
  cout << rH << endl;
  ofstream ofile("output/testHiggs.dat");
  ofile << as << "  " << rH << endl;
  for(as=0.04; as<0.29; as+=0.01) {
    rH = x*sxD.deltaHggAux(as, x, 125, 173, -4.7);
    ofile << as << "  " << rH << endl;
  }
  ofile.close();
  exit(0);
  //*/

  as = 0.215;
  double x_min = 1e-9;
  double x_max = 0.1;
  for(int j=0; j<Nas; j++) {
    for(int i=0; i<Npoints; i++) {
      x = x_min * exp(i/(Npoints-1.)*log(x_max/x_min));
      HELL::sqmatrix<double> dP = sxD.DeltaP(as, x);
      double dKhg = sxD.deltaKhg (as, x, mQ);
      cout << "as = " << as
	   << "     x = "<< setw(6) << x
	   << "    xDeltaPgg = "<< setw(12) << x*dP.gg()
	   << "    xDeltaPqg = "<< setw(12) << x*dP.qg()
	   << "    xDeltaCLg = "<< setw(12) << x*sxD.deltaCLg (as, x)/nf
	   <<                      setw(12) << x*sxD.deltaMCLg(as, x, mQ)
	   << "    xDeltaC2g = "<< setw(12) << x*sxD.deltaC2g (as, x)/nf
	   <<                      setw(12) << x*(sxD.deltaMC2g(as, x, mQ) - dKhg)
	   << "    xDeltaKhg = "<< setw(12) << x*dKhg
	   << endl;
    }
    as -= 0.01;
    cout << endl;
  }


  // Finish time
  gettimeofday(&t1,NULL);
  double t=t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001;
  cout << "Total time: " << t << "s" << endl;

  return 0;

}
