#pragma once

#include <cmath>
#include <vector>
#include "math/matrix.hh"


namespace HELL {

  double phiprime(double xi1, double xi2, double phi); // angle between k1+k2 and k2
  double phinonprime(double xi, double xi2, double phiprime); // angle between k1 and k2


class QQbarPair {
public:
  QQbarPair(double mQ, double muFrat=1); // mQ = mass of heavy quark Q
  ~QQbarPair() { return; };
  void   SetFactorizationScale(double muFrat) { _muFrat=muFrat; };
  double GetFactorizationScale() { return _muFrat; };
  void   SetKinematics(double x, double invM, double rapY, double pt);
  double GetX() { return _x; };
  double GetInvariantMass() { return _invM; };
  double GetRapidity() { return _rapY; };
  double GetTransverseMomentum() { return _pt; };
  //double GetTransverseMomentum() { return (_pt>0 ? _pt : sqrt(1/_x-1)); };
  double GetQthat() { return _pt/_invM; };
  // this function returns a vector A such that h_{k1,k2} = A[k1+k2][k1], up to k1+k2 <= maxorder
  //std::vector<std::vector<double> > hCoeffs(int maxorder=3);
  //std::vector<std::vector<double> > hCoeffsSaved();
  //void initCoeffs(int maxorder=3);
  //double h1();
  //double factorH();
  //double h_kp_km(int kp, int km, double *err = NULL);
  //
  double diffXS_M_Y_pt     (double xi1, double xi2,     double intvar1, double intvar2);
  double diffXS_M_Y_pt_SUB (double xi1, double intvar0, double intvar1, double intvar2, double &xi2p);
  double diffXS_M_Y_pt_SUBx(double intvar0, double intvar1, double intvar2);
  //double diffXS_M_Y_pt_SUBc(double intvar1, double intvar2);
  //double diffXS_M_Y_pt_v2  (double xi2, double phip,    double intvar1, double intvar2);
  double diffXS_M_Y_pt_Aux (double intvar1, double intvar2);
  //
  double diffXS_M_Y    (double xi1, double xi2, double intvar1, double intvar2, double phi);
  double diffXS_M_Y_Aux(double intvar1, double intvar2, double xi);
  double diffXS_M_Y_LO (double intvar1, double intvar2, double intvar3);
  double diffXS_M_Y_LO_qqbarchannel(double intvar1);
  //
  friend double integrandGSLxxx1(double x, void *pars);
  friend double integrandGSLxxx2(double x, void *pars);
  double TESTzzdCoff_dphi_int(double xi1, double xi2, double omega, double beta);
  double TESTzzCoff_int(double xi1, double beta);
  //double zzCoff(double xi1, double omega, double beta);
  //double zzC(double beta);
  double intphi(double intvar, double xi1, double xi2, double &jac);
  double dCoff_dphi(double xi1, double xi2, double phi, double omega, double beta); // new
  double dCoff_dphi_0(double xi1, double omega, double beta);
  double dCoff_dphi_0(double xi1, double phi, double omega, double beta);
  double Coff(double xi1, double omega, double beta);
  double Con(double beta);
  double Con_qqbarchannel(double beta);
  //
private:
  double _mQ, _muFrat;
  double _x, _invM, _rapY, _pt;
  //bool areCoeffsSaved;
  //std::vector<std::vector<double> > savedCoeffs;
  //std::vector<dcomplex> Yi;
  //double zzdCoff_dphi_v1(double xi1, double xi2, double phi, double beta1, double beta2);
  //double zzdCoff_dphi(double xi1, double xi2, double phi, double omega, double beta);
};



class QQbarSingleQuark {
public:
  QQbarSingleQuark(double mQ, double muFrat=1); // mQ = mass of heavy quark Q
  ~QQbarSingleQuark() { return; };
  void   SetFactorizationScale(double muFrat) { _muFrat=muFrat; };
  double GetFactorizationScale() { return _muFrat; };
  void   SetKinematics(double x, double rapY, double pt); // variables at collinear coefficient function level
  double GetX() { return _x; };
  double GetRapidity() { return _rapY; };
  double GetTransverseMomentum() { return _pt; };
  double GetPthat() { return _pt/_mQ; };
  //
  double diffXS_Y_pt    (double xi1, double xi2, double z, double eta, double intvar1);
  double diffXS_Y_pt_v2 (double xi , double xi2, double z, double eta, double phip, double xi1);
  double diffXS_Y_pt_v3 (double xi1, double xi2, double z, double eta, double phi);
  double diffXS_Y_pt_Aux(double xi1, double z, double eta);
  double diffXS_Y_pt_Aux_v1(double xi1, double z, double eta);
  double diffXS_Y_pt_Aux_v2(double xi1, double z, double eta);
  double diffXS_Y_pt_Aux_v2_subplus(double z, double eta);
  double diffXS_Y_pt_LO(double z, double eta, double pthat2);
  double diffXS_Y_pt_LO_qqbarchannel(double z, double eta, double pthat2);
  //
  //double diffXS_Y    (double xi1, double xi2, double intvar1, double intvar2, double intvar3);
  //double diffXS_Y_Aux(double intvar1, double intvar2);
  //
  //double intphi(double intvar, double xi1, double xi2, double &jac);
  //
private:
  double _mQ, _muFrat;
  double _x, _rapY, _pt;
  //double dCoff_dphi(double xi1, double xi2, double phi, double omega, double beta); // new
  //double Coff(double xi1, double omega, double beta); // new
};


};

