#pragma once
#include <cmath>


namespace HELL {

  extern bool useLLp;

  enum process {
    U        =  0,
    dU       =  10,
    Uqg      =  1,
    gamma_qg =  2,
    C2       =  3,
    CL       =  4,
    MSbarC2  =  5,
    MSbarCL  =  6,
    // mass dependent coefficients are >50
    Khg   = 51,
    MC2   = 62,
    MCL   = 63,
    MC2ax = 64,
    MCLax = 65,
    MC2cc = 66,
    MCLcc = 67,
    MC3cc = 68,
    // Higgs
    ggH    = 101,
    ggHaux = 102,
    // Drell-Yan
    DY = 201,
    // QQbar production
    check0 = 3000,
    check1 = 3001,
    check1SUB = 3002,
    QQbarPairM2Yq_part1 = 301,
    QQbarPairM2Yq_part2 = 302,
    QQbarPairM2Yq    = 303,
    QQbarPairM2YqAux = 304,
    QQbarPairM2Yq_SUB1 = 3011,
    QQbarPairM2Yq_SUB4 = 3012,
    QQbarPairM2Yq_SUBc = 3013,
    QQbarPairM2Y     = 321,
    QQbarPairM2Y_part1 = 3211,
    QQbarPairM2YAux  = 322,
    QQbarPairM2YAux_part2  = 3221,
    QQbarPairM2YLO   = 323,
    QQbarPairM2YLO_qqbarchannel   = 3231,
    QQbarPairM2Y_v2  = 324,
    QQbarPairM2Yq_v2 = 325,
    QQbarSingleQuarkYq    = 311,
    QQbarSingleQuarkYqAux = 312,
    QQbarSingleQuarkYqAux_v2 = 3121,
    QQbarSingleQuarkY     = 313,
    QQbarSingleQuarkYAux  = 314,
    QQbarSingleQuarkYq_v2 = 316,
    // scheme change function
    R    = -1,
    logR = -2,
    // test
    Ctest1 = -1001,
    Ctest2 = -1002
  };

  class Interpolator;
  struct p2s { double mQ; Interpolator *i; };


  inline double Power(double a, double b) { return pow(a,b); }
  inline double Sqrt(double a) { return sqrt(a); }
  inline double Log(double a) { return log(a); }
  inline double ArcSinh(double a) { return log(a+sqrt(1+a*a)); }
  inline double ArcCsch(double a) { if(a<1e-4) return -log(a/2)+a*a/4; return log(1/a+sqrt(1+1/a/a)); }
  inline double ArcCoth(double a) { return (log(1+1./a)-log(1-1./a))/2.; }
  inline double ArcTanh(double a) { return (log(1+a)-log(1-a))/2.; }


  const double Nc = 3.;
  const double CA = Nc;
  const double TF = 0.5;
  const double CF = (Nc*Nc-1.)/2./Nc;
  inline double beta0(int nf) { return (11.*CA - 2.*nf)/12./M_PI; }


  const double ZETA2  =  1.64493406684822643647;
  const double ZETA3  =  1.2020569031595942855;
  const double ZETA4  =  1.082323233711138191516;
  const double ZETA5  =  1.036927755143369926331;
  const double EulerGamma = 0.5772156649015328606065120900824;

};
