#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <cstdlib>
#include <cuba.h>
#include <gsl/gsl_errno.h>

#include "CoefficientFunctionsMultiDiff.hh"
#include "QQbarPair.hh"
#include "Pqg.hh"
#include "math/borel.hh"
#include "math/special_functions.hh"
#include "integration.hh"
#include "version.hh"
#include "expansion_AD_SF.hh"
//#include <complex_bessel/sp_airy.h>
#include <gsl/gsl_sf.h>


using namespace std;



namespace HELL {

  extern double phinonprime(double xi, double xi2, double phiprime);


  namespace HELLdiff {


    double HTdamp(double xi, double asb0) {
      if(xi<=exp(-1/asb0)) return 0.;
      if(xi<1) {
	int kk = 1;//10;
	int jj = 1;//1000;
	double  HTdamp = 1 - pow(-asb0*log(xi),kk*(1+1/asb0/jj));
	return HTdamp;
      }
      return 1.;
    }
    double dHTdamp(double xi, double asb0) {
      if(xi<=exp(-1/asb0)) return 0.;
      if(xi<1) {
	int kk = 1;//10;
	int jj = 1;//1000;
	double  damp = HTdamp(xi,asb0);
	double Ddamp = (damp-1) * kk*(1+1/asb0/jj)/xi/log(xi);
	return Ddamp;
      }
      return 1.;
    }
    double approxUreg_nodamp(double xi, double x, double as, double b0, double pole, double residue, bool RC=true) {
      if(x ==1) return 0;
      if(xi==1) return 0;
      if(RC && xi<=exp(-1/as/b0)) return 0;
      double logxi = residue*log(xi);
      if(RC) logxi = residue*log(1+as*b0*log(xi))/as/b0;
      double log1x = -log(x);
      double BessArg, Ureg;
      if(logxi>=0) {
	BessArg = 2*sqrt(log1x*logxi);
	//cout << logxi << "  " << log1x << "  " << BessArg << endl;
	Ureg = gsl_sf_bessel_I1(BessArg)*sqrt(logxi/log1x)/x;
      } else {
	BessArg = 2*sqrt(-log1x*logxi);
	//cout << logxi << "  " << log1x << "  " << BessArg << endl;
	Ureg = -gsl_sf_bessel_J1(BessArg)*sqrt(-logxi/log1x)/x;
      }
      Ureg *= pow(x,-pole);
      return Ureg;
    }
    double approxUreg(double xi, double x, double as, double b0, double pole, double residue, bool RC=true) {
      double Ureg = approxUreg_nodamp(xi, x, as, b0, pole, residue, RC);
      double damp = (1-x)*(1-x) * HTdamp(xi, as*b0);
      return Ureg*damp;
    }
    double DapproxUreg_nodamp(double xi, double x, double as, double b0, double pole, double residue, bool RC=true) {
      if(x ==1) return 0;
      if(RC && xi<=exp(-1/as/b0)) return 0;
      double  logxi = residue*log(xi);
      double Dlogxi = residue/xi;
      if(xi==1) return Dlogxi/x;
      if(RC) {
	logxi  = residue*log(1+as*b0*log(xi))/as/b0;
	Dlogxi = residue/(1+as*b0*log(xi))/xi;
      }
      double log1x = -log(x);
      double BessArg, Ureg;
      if(logxi>=0) {
	BessArg = 2*sqrt(log1x*logxi);
	//cout << logxi << "  " << log1x << "  " << BessArg << endl;
	Ureg = Dlogxi*(2*gsl_sf_bessel_I1(BessArg)/BessArg + gsl_sf_bessel_I0(BessArg) + gsl_sf_bessel_In(2,BessArg))/x/2;
      } else {
	BessArg = 2*sqrt(-log1x*logxi);
	//cout << logxi << "  " << log1x << "  " << BessArg << endl;
	Ureg = Dlogxi*(2*gsl_sf_bessel_J1(BessArg)/BessArg + gsl_sf_bessel_J0(BessArg) - gsl_sf_bessel_Jn(2,BessArg))/x/2;
      }
      Ureg *= pow(x,-pole);
      return Ureg;
    }
    double DapproxUreg(double xi, double x, double as, double b0, double pole, double residue, bool RC=true) {
      double  Ureg =  approxUreg_nodamp(xi, x, as, b0, pole, residue, RC);
      double dUreg = DapproxUreg_nodamp(xi, x, as, b0, pole, residue, RC);
      double  damp = (1-x)*(1-x) *  HTdamp(xi, as*b0);
      double Ddamp = (1-x)*(1-x) * dHTdamp(xi, as*b0);
      return Ureg*Ddamp + dUreg*damp;
    }






    void checkDimension(int integraldim, int integranddim) {
      if(integraldim < integranddim) {
	cout << "Error: integral dimension " << integraldim << " is insufficient: " << integranddim << endl;
        exit(78);
      } else if(integraldim > integranddim) {
	cout << "Warning: integral dimension " << integraldim << " is larger than needed: " << integranddim << endl;
      }
      return;
    }



    // QQbar Pair, triple differential
    double intQQbarPairM2Yq(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 4);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xiF = muFrat*muFrat;
      //
      double logxi0 = log(xiF)-1./asb0;
      double xi0 = exp(logxi0);
      //double xi0 = xiF*exp(-1./asb0);
      double mult=1.2;
      //*
      //double tmin=qthat*qthat/4;  double tmax=min(max(xiF*pow(xiF/xi0,mult),100*qthat*qthat),1e3);
      double tmin=max(qthat*qthat/4,xi0);  double tmax=min(1e3,max(xiF*pow(xiF/xi0,mult),100*qthat*qthat));
      //double u=intVar[0]*log(tmax/tmin);    double t1=tmin*exp(u);     jac*=t1*log(tmax/tmin);
      //double t1=exp((1-intVar[0])*log(tmin)+intVar[0]*log(tmax));     jac*=t1*log(tmax/tmin);
      //double y1max=min(qthat/2,sqrt(t1)-sqrt(xi0)); double y1=y1max*intVar[1]; jac*=y1max;
      //double u=intVar[0]*log(tmax/tmin);    double t1=tmin*exp(u);     jac*=t1*log(tmax/tmin);
      //double t1=exp((1-intVar[0])*log(tmin)+intVar[0]*log(tmax));     jac*=t1*(log(tmax)-log(tmin));
      double u=intVar[0]*intVar[0]; jac*=2*intVar[0]; double t1=exp((1-u)*log(tmin)+u*log(tmax));     jac*=t1*(log(tmax)-log(tmin));
      //double u=intVar[0]*intVar[0]*intVar[0]; jac*=3*intVar[0]*intVar[0]; double t1=exp((1-u)*log(tmin)+u*log(tmax));     jac*=t1*(log(tmax)-log(tmin));
      double y1max=min(qthat/2,sqrt(t1)-exp(logxi0/2)); double y1=y1max*(1-intVar[1]*intVar[1]); jac*=2*intVar[1]*y1max;
      //double y1max=min(qthat/2,sqrt(t1)-exp(logxi0/2)); double y1=y1max*(1-intVar[1]*intVar[1]*intVar[1]*intVar[1]); jac*=4*intVar[1]*intVar[1]*intVar[1]*y1max;
      /*/
      double y1max=qthat/2; double y1=y1max*(1-intVar[1]); jac*=y1max;
      double tmin=max(qthat*qthat/4,pow(exp(logxi0/2)+y1,2));  double tmax=min(max(xiF*pow(xiF,mult)*exp(-mult*logxi0),100*qthat*qthat),1e5);
      double u=intVar[0]*log(tmax/tmin);  //jac *= 2*intVar[0];
      double t1=tmin*exp(u);     jac*=t1*log(tmax/tmin);
      //*/
      //double t2=xiF*exp(mult*u); jac2*=mult*t2*u0; double y2max=min(qthat/2,sqrt(t2)-sqrt(xi0)); double y2=y2max*intVar[1]; jac2*=y2max;
      //xi1a = t1*pow(1+y1/sqrt(t1),2);  xi2a = t1*pow(1-y1/sqrt(t1),2);  jac*=4*sqrt(t1)*(1-y1*y1/t1);
      xi1 = pow(sqrt(t1)+y1,2);  xi2 = pow(sqrt(t1)-y1,2);  jac*=4*(sqrt(t1)-y1*y1/sqrt(t1));
      //xi1a = t1+y1*y1+2*y1*sqrt(t1);  xi2a = t1+y1*y1-2*y1*sqrt(t1);  jac*=4*sqrt(t1)*(1-y1*y1/t1);
      //if(xi2a<xi0) cout << y1max << "   " << y1 << "   " << xi1a << "   " << xi2a << "   " << xi0 << endl;
      //xi1b = t2*pow(1+y2/sqrt(t2),2);  xi2b = t2*pow(1-y2/sqrt(t2),2);  jac2*=4*sqrt(t2)*(1-y2*y2/t2);
      //double sqrtarg = 4*xi1a*xi2a-(qthat*qthat-xi1a-xi2a)*(qthat*qthat-xi1a-xi2a);
      //if(sqrtarg<0) {
      //  cout << ">>>>>Warning: argument of sqrt is negative: " << sqrtarg << endl;
      //}
      //
      //double ximin=xi0;  double ximax=min(xiF*pow(xiF/xi0,mult),1e3);
      //xi1 = exp((1-intVar[0])*log(ximin)+intVar[0]*log(ximax));     jac*=xi1*(log(ximax)-log(ximin));
      //xi2 = exp((1-intVar[1])*log(ximin)+intVar[1]*log(ximax));     jac*=xi2*(log(ximax)-log(ximin));
      //
      /*
      // reduces integration range to xi_{1,2} >= 1, for direct comparison with SUB4
      jac=1;
      xi0=1e-6; //xiF*exp(-1./a0/b0);
      tmin=max(qthat*qthat/4,1.);  tmax=min(1e3,max(xiF*pow(xiF/xi0,mult),100*qthat*qthat));
      u=intVar[0]*log(tmax/tmin);
      double t=tmin*exp(u);     jac*=t*log(tmax/tmin);      double ymax=min(qthat/2,sqrt(t)-1.); double y=ymax*intVar[1]; jac*=ymax;
      xi1 = t*pow(1+y/sqrt(t),2);  xi2 = t*pow(1-y/sqrt(t),2);  jac*=4*sqrt(t)*(1-y*y/t);
      //*/
      /*
      // reduces integration range to xi_{1,2} >= 1e-6
      jac=1;
      xi0=exp(logxi0);
      tmin=max(qthat*qthat/4,xi0);  tmax=min(1e3,max(xiF*pow(xiF/1e-6,mult),100*qthat*qthat));
      u=intVar[0]*log(tmax/tmin);
      double t=tmin*exp(u);     jac*=t*log(tmax/tmin);      double ymax=min(qthat/2,sqrt(t)-sqrt(xi0)); double y=ymax*intVar[1]; jac*=ymax;
      xi1 = t*pow(1+y/sqrt(t),2);  xi2 = t*pow(1-y/sqrt(t),2);  jac*=4*sqrt(t)*(1-y*y/t);
      //*/
      //
      double sqrtx = sqrt(x*(1+qthat*qthat));
      x1  = sqrtx*exp( rapY);
      x2  = sqrtx*exp(-rapY);
      if(x1>1 || x2>1) return 0;
      //
      double cosphi = (qthat*qthat-xi1-xi2)/2./sqrt(xi1*xi2);
      if(fabs(cosphi)>1.) {
	//cout << "theta function in action: |cos(phi)|-1 = " << fabs(cosphi)-1 << "  xi1="<<xi1 << "  xi2="<<xi2 << "  qthat="<<qthat << endl;
	return 0;
      }
      //cout << "xi0=" << xi0 << "   " << xi1 << "  " << xi2 << "  " << t1 << "  " << y1 << "  " << intVar[0] << "  " << intVar[1] << endl;
      return fabs(jac) * QQbar->diffXS_M_Y_pt(xi1, xi2, intVar[2], intVar[3]);
    }
    /*
    double intQQbarPairM2Yq_v2(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 4);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xrescale = sqrt(1+qthat*qthat);
      double xiF = muFrat*muFrat;
      //
      double xi0=xiF*exp(-1./asb0); double mult=2;
      double tmin=xi0;  double tmax=min(1e3,xiF*pow(xiF/xi0,mult));
      double u=intVar[0]*log(tmax/tmin);
      double t1=tmin*exp(u);     jac*=t1*log(tmax/tmin);
      double cosphipmax = (t1+qthat*qthat-xi0)/2./qthat/sqrt(t1);
      double phipmin = (cosphipmax>1? 0 : myacos(cosphipmax));
      double phip = phipmin + 2*(M_PI-phipmin)*intVar[1];   jac*=2*(M_PI-phipmin);
      xi1 = t1+qthat*qthat-2*qthat*sqrt(t1)*cos(phip);  xi2 = t1;
      //
      x1  = sqrt(x)*exp( rapY) * xrescale;
      x2  = sqrt(x)*exp(-rapY) * xrescale;
      return fabs(jac) * QQbar->diffXS_M_Y_pt_v2(xi2, phip, intVar[2], intVar[3]);
    }
    */
    double intQQbarPairM2Yq_part1(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par, double &xi2s, double &sub, bool issub=false) {
      checkDimension(dim, 4);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xrescale = sqrt(1+qthat*qthat);
      double xiF = muFrat*muFrat;
      double xi0 = xiF*exp(-1./asb0);
      if(xi0<1e-6 || issub) xi0 = 1e-6;
      //xi0 *= 5;
      //double u = intVar[0];
      //double u = intVar[0]*intVar[0]; jac *= 2*intVar[0];
      double u = intVar[0]*intVar[0]*intVar[0]; jac *= 3*intVar[0]*intVar[0];
      //double u = exp(-1/intVar[0]); jac *= u;
      double phi = 2*M_PI*intVar[3];
      double xi1max = qthat*qthat/2./(1+cos(phi));
      if(xi1max<xi0) xi1max = xi0;
      if(xi1max>1/xi0) xi1max = 1/xi0;
      //if(xi1max>1) xi1max = 1;
      xi1 = xi0+(xi1max-xi0)*u;  xi2 = 0;  jac*=xi1max-xi0;
      x1  = sqrt(x)*exp( rapY) * xrescale;
      x2  = sqrt(x)*exp(-rapY) * xrescale;
      double res = fabs(jac) * QQbar->diffXS_M_Y_pt_SUB(xi1, intVar[1], intVar[2], intVar[3], xi2);
      sub = fabs(jac) * QQbar->diffXS_M_Y_pt_SUBx(intVar[1], intVar[2], intVar[3]);
      xi2s = qthat*qthat;
      return res;
    }
    double intQQbarPairM2Yq_part2(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par, bool issub=false) {
      checkDimension(dim, 4);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xrescale = sqrt(1+qthat*qthat);
      double xiF = muFrat*muFrat;
      double xi0 = xiF*exp(-1./asb0);
      if(xi0<1e-6 || issub) xi0 = 1e-6;
      //xi0 = 0.01;
      //xi0 *= 5;
      //double u = intVar[0];
      double u = intVar[0]*intVar[0]; jac *= 2*intVar[0];
      //double u = intVar[0]*intVar[0]*intVar[0]; jac *= 3*intVar[0]*intVar[0];
      //double u = exp(-1/intVar[0]); jac *= u;
      double phi = 2*M_PI*intVar[3];
      double xi1max = qthat*qthat/2./(1+cos(phi));
      if(xi1max<xi0) xi1max = xi0;
      if(xi1max>1/xi0) xi1max = 1/xi0;
      //xi1 = xiF + (xi1max-xiF)*u;  xi2 = qthat*qthat;  jac*=xi1max-xiF;
      if(xi1max>xiF) { xi1 = xiF + (xi1max-xiF)*u;  xi2 = qthat*qthat;  jac*=xi1max-xiF; }
      else           { xi1 = xi1max + (xiF-xi1max)*u;  xi2 = qthat*qthat;  jac*=xi1max-xiF; }
      //xi1 = xi0 + (xi1max-xi0)*u;  xi2 = qthat*qthat;  jac*=xi1max-xi0; // using this does not change the resummed result, as the integral of U' from 0 to 1 is zero (checked explictly also here)
      x1  = sqrt(x)*exp( rapY) * xrescale;
      x2  = sqrt(x)*exp(-rapY) * xrescale;
      // here we shall not use fabs(jac) because jac contains the sign of the integral
      return jac * QQbar->diffXS_M_Y_pt_SUBx(intVar[1], intVar[2], intVar[3]);
    }
    double intQQbarPairM2Yq_Aux(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 2);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      //
      xi1 = qthat*qthat;
      x1  = x * (1+xi1);
      return fabs(jac) * QQbar->diffXS_M_Y_pt_Aux(intVar[0], intVar[1]);
    }



    // QQbar Pair, double differential
    double intQQbarPairM2Y_v2(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 5);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double xiF = muFrat*muFrat;
      //
      //double qthat = intVar[4]/(1-intVar[4]);
      double qtmax = sqrt(exp(-2*fabs(QQbar->GetRapidity()))/x - 1);
      double qthat = intVar[4]*qtmax;
      QQbar->SetKinematics(QQbar->GetX(), QQbar->GetInvariantMass(), QQbar->GetRapidity(), qthat*QQbar->GetInvariantMass());
      double xi0=xiF*exp(-1./asb0); double mult=1.2;
      double tmin=max(qthat*qthat/4,xi0);  double tmax=max(xiF*pow(xiF/xi0,mult),100*qthat*qthat);
      double u=intVar[0]*log(tmax/tmin);
      double t1=tmin*exp(u);     jac*=t1*log(tmax/tmin);      double y1max=min(qthat/2,sqrt(t1)-sqrt(xi0));  double y1=y1max*intVar[1];  jac*=y1max;
      xi1 = t1*pow(1+y1/sqrt(t1),2);  xi2 = t1*pow(1-y1/sqrt(t1),2);  jac *= 4*sqrt(t1)*(1-y1*y1/t1);
      //jac *= 2*qthat/(1-intVar[4])/(1-intVar[4]);
      jac *= 2*qthat*qtmax;
      //
      double xrescale = sqrt(1+qthat*qthat);
      x1  = sqrt(x)*exp( rapY) * xrescale;
      x2  = sqrt(x)*exp(-rapY) * xrescale;
      return fabs(jac) * QQbar->diffXS_M_Y_pt(xi1, xi2, intVar[2], intVar[3]);
    }
    double intQQbarPairM2Y(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 5);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double xiF = muFrat*muFrat;
      /*
      double logxi0 = log(xiF)-1./asb0;
      double xi0 = exp(logxi0);
      double mult=1.2;
      double tmin=xi0;  double tmax=min(xiF*pow(xiF/xi0,mult),1e5);
      double t1=exp((1-intVar[0])*log(tmin)+intVar[0]*log(tmax));     jac1*=t1*log(tmax/tmin);
      double y1max=sqrt(t1)-exp(logxi0/2); double y1=y1max*(1-intVar[1]*intVar[1]); jac1*=2*intVar[1]*y1max;
      xi1a = pow(sqrt(t1)+y1,2);  xi2a = pow(sqrt(t1)-y1,2);  jac1*=4*sqrt(t1)*(1-y1*y1/t1);
      /*/
      double logxi0 = log(xiF)-1./asb0;
      double xi0 = exp(logxi0);
      double mult=1.2;
      double tmin=xi0;  double tmax=min(xiF*pow(xiF/xi0,mult),1e5);
      if(tmax < tmin) cout << "What??" << endl;
      xi1 = exp((1-intVar[0])*log(tmin)+intVar[0]*log(tmax));  jac*=xi1*log(tmax/tmin);
      xi2 = exp((1-intVar[1])*log(tmin)+intVar[1]*log(tmax));  jac*=xi2*log(tmax/tmin);
      //*/
      //
      double jacphi, phi = QQbar->intphi(intVar[4], xi1, xi2, jacphi);   jac *= jacphi;
      double xrescale = sqrt(1+xi1+xi2+2*sqrt(xi1*xi2)*cos(phi));
      x1  = sqrt(x)*exp( rapY) * xrescale;
      x2  = sqrt(x)*exp(-rapY) * xrescale;
      return fabs(jac) * QQbar->diffXS_M_Y(xi1, xi2, intVar[2], intVar[3], phi);
    }
    double intQQbarPairM2Y_part1(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 5);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double xiF = muFrat*muFrat;
      /*
      double logxi0 = log(xiF)-1./asb0;
      double xi0 = exp(logxi0);
      double mult=1.2;
      double tmin=xi0;  double tmax=min(xiF*pow(xiF/xi0,mult),1e5);
      double t1=exp((1-intVar[0])*log(tmin)+intVar[0]*log(tmax));     jac1*=t1*log(tmax/tmin);
      double y1max=sqrt(t1)-exp(logxi0/2); double y1=y1max*(1-intVar[1]*intVar[1]); jac1*=2*intVar[1]*y1max;
      xi1a = pow(sqrt(t1)+y1,2);  xi2a = pow(sqrt(t1)-y1,2);  jac1*=4*sqrt(t1)*(1-y1*y1/t1);
      /*/
      double logxi0 = log(xiF)-1./asb0;
      double xi0 = exp(logxi0);
      double mult=3;
      //double tmin=xi0;  double tmax=min(xiF*pow(xiF/xi0,mult),1e5);
      //if(tmax < tmin) cout << "What??" << endl;
      //xi1 = exp((1-intVar[0])*log(tmin)+intVar[0]*log(tmax));  xi2 = exp((1-intVar[1])*log(xi1)+intVar[1]*log(tmax));  jac*=xi1*xi2*log(tmax/tmin)*log(tmax/xi1);
      double qthat2max = exp(-2*fabs(QQbar->GetRapidity()))/QQbar->GetX() - 1;
      double phi = 2*M_PI*intVar[4];  jac *= 2*M_PI;
      double cosphi = cos(phi);
      double xi1max = qthat2max/2./(1.+cosphi);
      if(xi1max<xi0) xi1max = xi0;
      if(xi1max>mult/xi0) xi1max = mult/xi0;
      //double u1 = intVar[0]*intVar[0]*intVar[0]*intVar[0]; jac *= 4*intVar[0]*intVar[0]*intVar[0];
      //xi1 = xi0+(xi1max-xi0)*u1;  jac*=xi1max-xi0;
      xi1 = exp((1-intVar[0])*logxi0+intVar[0]*log(xi1max));  jac*=xi1*(log(xi1max)-logxi0);
      if(xi1>xi1max) cout << "wtf???    " << xi1 << "  " << xi1max << "  " << cosphi << endl;
      double sqrtxi2p = -sqrt(xi1)*cosphi + sqrt(qthat2max-xi1*(1-cosphi*cosphi));
      double xi2max = sqrtxi2p*sqrtxi2p;
      if(xi2max<xi1) cout << "wtf???x   " << xi1 << "  " << xi2max << "  " << cosphi << "  " << xi1max << "  " << qthat2max-xi1*(1-cosphi*cosphi) << endl;
      //double u2 = intVar[1]*intVar[1]*intVar[1]; jac *= 3*intVar[1]*intVar[1];
      //xi2 = xi1+(sqrtxi2p*sqrtxi2p-xi1)*u2;  jac*=sqrtxi2p*sqrtxi2p-xi1;
      xi2 = exp((1-intVar[1])*log(xi1)+intVar[1]*log(xi2max));  jac*=xi2*(log(xi2max)-log(xi1));
      if(xi2<xi1) cout << "wtf???xx  " << xi1 << "  " << xi2 << "  " << cosphi << endl;
      //*/
      //
      double xrescale = sqrt(1+xi1+xi2+2*sqrt(xi1*xi2)*cosphi);
      x1  = sqrt(x)*exp( rapY) * xrescale;
      x2  = sqrt(x)*exp(-rapY) * xrescale;
      //return fabs(jac) * QQbar->diffXS_M_Y(xi1, xi2, intVar[2], intVar[3], phi);
      return fabs(jac) * (QQbar->diffXS_M_Y(xi1, xi2, intVar[2], intVar[3], phi) + QQbar->diffXS_M_Y(xi1, xi2, intVar[2], intVar[3], -phi)) /2.;
      //return fabs(jac) * (QQbar->diffXS_M_Y(xi1, xi2, intVar[2], intVar[3], phi) + QQbar->diffXS_M_Y(xi2, xi1, intVar[2], intVar[3], phi)) /2.;
    }
    double intQQbarPairM2YAux(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 2);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      //
      x1  = exp(-2*rapY);
      xi1 = exp(-2*rapY)/x-1;
      //if(x1>1) return 0;
      return fabs(jac) * 2 * QQbar->diffXS_M_Y_Aux(intVar[0], intVar[1], xi1);
    }
    double intQQbarPairM2YAux_part2(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 2);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      //
      x1  = x;
      xi1 = exp(-2*rapY)/x-1;
      //if(x1>1) return 0;
      return fabs(jac) * 2 * QQbar->diffXS_M_Y_LO(intVar[0], intVar[1], 0) * (1+xi1);
    }

    double intQQbarPairM2YLO(int dim, double *intVar, void *par) {
      checkDimension(dim, 3);
      QQbarPair *QQbar = (QQbarPair*) par;
      return QQbar->diffXS_M_Y_LO(intVar[0], intVar[1], intVar[2]);
    }
    double intQQbarPairM2YLO_qqbarchannel(int dim, double *intVar, void *par) {
      checkDimension(dim, 1);
      QQbarPair *QQbar = (QQbarPair*) par;
      return QQbar->diffXS_M_Y_LO_qqbarchannel(intVar[0]);
    }
    /*
    double intQQbarPairM2Yq_SUB4(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, void *par) {
      checkDimension(dim, 4);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xrescale = sqrt(1+qthat*qthat);
      double xiF = muFrat*muFrat;
      double xi0=1e-6; //xiF*exp(-1./a0/b0);
      double mult=1.2;
      double tmin=max(qthat*qthat/4,1.);  double tmax=min(1e3,max(xiF*pow(xiF/xi0,mult),100*qthat*qthat));
      double u=intVar[0]*log(tmax/tmin);
      double t=tmin*exp(u);     jac*=t*log(tmax/tmin);      double ymax=min(qthat/2,sqrt(t)-1.); double y=ymax*intVar[1]; jac*=ymax;
      xi1 = t*pow(1+y/sqrt(t),2);  xi2 = t*pow(1-y/sqrt(t),2);  jac*=4*sqrt(t)*(1-y*y/t);
      x1  = sqrt(x)*exp( rapY) * xrescale;
      x2  = sqrt(x)*exp(-rapY) * xrescale;
      return fabs(jac) * QQbar->diffXS_M_Y_pt(xi1, xi2, intVar[2], intVar[3]);
    }
    double intQQbarPairM2Yq_SUB1(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, void *par) {
      checkDimension(dim, 4);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xrescale = sqrt(1+qthat*qthat);
      //double xiF = muFrat*muFrat;
      double xi0=1e-6; //xiF*exp(-1./a0/b0);
      //xi0 *= 5;
      //double u = intVar[0];
      double u = intVar[0]*intVar[0]; jac *= 2*intVar[0];
      //double u = intVar[0]*intVar[0]*intVar[0]; jac *= 3*intVar[0]*intVar[0];
      //double u = exp(-1/intVar[0]); jac *= u;
      double phi = 2*M_PI*intVar[3];
      double xi1max = qthat*qthat/2./(1+cos(phi));
      //xi0 = 0.01;
      if(xi1max<xi0) xi1max = xi0;
      if(xi1max>1/xi0) xi1max = 1/xi0;
      //if(xi1max>1) xi1max = 1;
      xi1 = xi0+(xi1max-xi0)*u;  xi2 = 0;  jac*=xi1max-xi0;
      //xi1 = xi0+(1-xi0)*u; jac*=1-xi0;
      x1  = sqrt(x)*exp( rapY) * xrescale;
      x2  = sqrt(x)*exp(-rapY) * xrescale;
      return fabs(jac) * QQbar->diffXS_M_Y_pt_SUB(xi1, intVar[1], intVar[2], intVar[3], xi2);
      //return fabs(jac) * QQbar->diffXS_M_Y_pt_SUB(xi1, intVar[1], intVar[2], intVar[3], false, xi2);
    }
    double intQQbarPairM2Yq_SUBc(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, void *par) {
      checkDimension(dim, 2);
      double jac = 1;
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xrescale = sqrt(1+qthat*qthat);
      xi1 = 0;  xi2 = 0;
      x1  = sqrt(x)*exp( rapY) * xrescale;
      x2  = sqrt(x)*exp(-rapY) * xrescale;
      return fabs(jac) * QQbar->diffXS_M_Y_pt_SUBc(intVar[0], intVar[1]);
    }
    */




    // Single quark double differential Y-pt
    double intQQbarSingleQuarkYq(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 5);
      double jac = 1;
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double xiF = muFrat*muFrat;
      //
      double pthat = QQbar->GetPthat();
      double intz = x+(1-x)*intVar[3];
      rapY  = -log(x/intz) * (intVar[2]-1./2.); // this is yhat, the integration variable
      jac *= -log(x/intz) * (1-x);
      double xz = x/intz;
      double eta = QQbar->GetRapidity() - rapY;
      double sqrtstuff = sqrt(pthat*pthat + 1/intz - sqrt(1+pthat*pthat)*(exp(eta)+exp(-eta))/sqrt(intz));
      double qthatmax = sqrtstuff + pthat;
      double qthatmin = max( sqrtstuff - pthat, 0.);
      double xi0=xiF*exp(-1./asb0); double mult=1.2;
      double tmin=max(qthatmin*qthatmin/4,xi0);  double tmax=max(xiF*pow(xiF/xi0,mult),100*qthatmin*qthatmin);
      double u=intVar[0]*log(tmax/tmin);
      double t1=tmin*exp(u);     jac*=t1*log(tmax/tmin);
      double y1max=min(qthatmax/2,sqrt(t1)-sqrt(xi0)); double y1=y1max*intVar[1]; jac*=y1max;
      xi1 = t1*pow(1+y1/sqrt(t1),2);  xi2 = t1*pow(1-y1/sqrt(t1),2);  jac*=4*sqrt(t1)*(1-y1*y1/t1);
      //
      x1  = sqrt(xz)*exp( rapY);
      x2  = sqrt(xz)*exp(-rapY);
      return fabs(jac) * QQbar->diffXS_Y_pt(xi1, xi2, intz, eta, intVar[4]) /intz;
    }
    // double intQQbarSingleQuarkYq_v2(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
    //   checkDimension(dim, 5);
    //   double jac = 1;
    //   QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
    //   double xiF = muFrat*muFrat;
    //   double pthat = QQbar->GetPthat();
    //   //
    //   // y,   x = partonic on-shell, arguments of C
    //   // eta, z = partonic off-shell, arguments of {\cal C}
    //   // etabar = y-eta
    //   // x/z         <= exp(-2|etabar|)   <<------ arguments of U <= 1
    //   // z(1+pthat2) <= exp(-2|eta|)      <<------ z1,z2 <= 1
    //   // prodotto:
    //   // A^2 <= exp(-2|eta|-2|etabar|) <= exp(-2|eta+etabar|) = exp(-2|y|)
    //   // from the first inequality we get:  A e^y <= exp(2etabar) <= e^y/A
    //   // from the condition X >= 2B (see paper for the notation), we get a more stringent condition:
    //   // A*exp(y)/(1-A*exp(-y)) <= exp(2etabar) <= (1-A*exp(y))/(A*exp(-y))
    //   //
    //   //
    //   // integration in etabar
    //   //
    //   double y = rapY;  // this is y, the argument of the on-shell coefficient
    //   //cout << y << endl;
    //   //double exp2etabarmax = min( 1./xx, (1-exp(y)*sqrt(xx*(1+pthat*pthat)))/(-xx*pthat*pthat+exp(-y)*sqrt(xx*(1+pthat*pthat))) );
    //   //double exp2etabarmin = max( xx, (-xx*pthat*pthat+exp(y)*sqrt(xx*(1+pthat*pthat)))/(1-exp(-y)*sqrt(xx*(1+pthat*pthat))) );
    //   if((-x*pthat*pthat+exp(-y)*sqrt(x*(1+pthat*pthat)))<0) cout << "AAAAAAAAA" << endl;  // what are these?
    //   if((-x*pthat*pthat+exp( y)*sqrt(x*(1+pthat*pthat)))<0) cout << "BBBBBBBBB" << endl;
    //   double A = sqrt(x*(1+pthat*pthat));
    //   //double exp2etabarmax = (1-exp(y)*sqrt(xx*(1+pthat*pthat)))/(-xx*pthat*pthat+exp(-y)*sqrt(xx*(1+pthat*pthat)));
    //   //double exp2etabarmin = (-xx*pthat*pthat+exp(y)*sqrt(xx*(1+pthat*pthat)))/(1-exp(-y)*sqrt(xx*(1+pthat*pthat)));
    //   double exp2etabarmax = exp(y)*(1-A*exp(y))/A;
    //   double exp2etabarmin = exp(y)*A/(1-A*exp(-y));
    //   //cout << exp2etabarmax << "  " << exp2etabarmin << endl;
    //   if(exp2etabarmax < exp2etabarmin) cout << "Problem: etabar limits inverted" << endl;
    //   double etabar = 0.5*log(exp2etabarmin) + 0.5*(log(exp2etabarmax)-log(exp2etabarmin))*intVar[2]; // this is etabar, the integration variable
    //   //etabar = -1.69;
    //   //cout << etabar << endl;
    //   jac *= 0.5*(log(exp2etabarmax)-log(exp2etabarmin));
    //   //
    //   //
    //   // integration in z
    //   //
    //   double eta = y - etabar;  // this is eta = y-etabar, the argument of the off-shell coefficient
    //   double B = sqrt(1+pthat*pthat) * cosh(eta);
    //   double zmin = x*exp(2*fabs(etabar));  // comes from  x/z <= exp(-2|etabar|)
    //   //double zmax = 1./pow(B+sqrt(B*B-pthat*pthat),2);
    //   double zmax = 1./4./B/B; // new limit found on board
    //   if(zmax < zmin) {
    // 	if(zmin-zmax>1e-10) cout << "Problem: z limits inverted: " << zmin << " < " << zmax << "   difference: " << zmin-zmax << endl;
    // 	else zmin = zmax;
    //   }
    //   double z = zmin+(zmax-zmin)*(1+cos(M_PI*intVar[3]))/2;  // this is z, argument of {\cal C}
    //   //cout << "z = " << z << endl;
    //   jac *= (zmax-zmin)*M_PI*sin(M_PI*intVar[3])/2;
    //   //
    //   //
    //   // integration in xi, xi2
    //   //
    //   double stuff = pthat*pthat + 1/z - 2*B/sqrt(z);
    //   double sqrtstuff = sqrt(stuff);
    //   if(stuff<0 && stuff>-1e-10) sqrtstuff = 0;
    //   if(sqrtstuff != sqrtstuff) cout << "NaN again... " << stuff << endl;
    //   //cout << pthat*pthat + 1/z - sqrt(1+pthat*pthat)*(exp(eta)+exp(-eta))/sqrt(z) << endl;
    //   double sqrtximax = sqrtstuff + pthat; //min( sqrtstuff + pthat, sqrt(1./z-4));
    //   double sqrtximin = sqrtstuff - pthat;
    //   if(sqrtximin < 0) { cout << "sqrtxi_min = " << sqrtximin << " < 0" << endl;   sqrtximin = 0; }
    //   if(sqrtximax < sqrtximin) cout << "Problem: xi limits inverted" << endl;
    //   double xi0=xiF*exp(-1./asb0); double mult=2;//0.2;
    //   double ximin=sqrtximin*sqrtximin;  double ximax=sqrtximax*sqrtximax;
    //   double x0 = intVar[0], epsilon = 1e-5;
    //   if(intVar[0]<  epsilon) x0 =   epsilon;
    //   if(intVar[0]>1-epsilon) x0 = 1-epsilon;
    //   //double u1=intVar[0]*log(t1max/t1min);
    //   //double t1=t1min*exp(u1);     jac*=t1*log(t1max/t1min);  // xi
    //   //double t1=t1min + (t1max-t1min)*x0;     jac*=(t1max-t1min);  // xi
    //   double xi = ximin + (ximax-ximin)*(1+cos(M_PI*x0))/2;     jac*=(ximax-ximin)*M_PI*sin(M_PI*x0)/2;  // xi
    //   //if(t1 != t1) cout << "NaN ...  " << t1max << "  " << t1min << "  " << t1 << endl;
    //   double xi2min=xi0;  double xi2max=xiF*pow(xiF/xi0,mult);
    //   //double t2=t2min + (t2max-t2min)*intVar[1];     jac*=(t2max-t2min); // xi2
    //   //double u2=intVar[1]*log(t2max/t2min);     double t2=t2min*exp(u2);     jac*=t2*log(t2max/t2min); // xi2
    //   double yxi=max(1e-10,log(xi/xi2min)/log(xi2max/xi2min));  double yxi2=pow(yxi*pow(1-yxi,2),1./3.);  double axi=yxi-yxi*(1-yxi)/yxi2+yxi2;  double y2 = (pow(intVar[1]*intVar[1]-axi,3)+pow(axi,3))/(pow(1-axi,3)+pow(axi,3));
    //   double u2=y2*log(xi2max/xi2min);     xi2=xi2min*exp(u2);     jac*=xi2*log(xi2max/xi2min)*3*pow(intVar[1]*intVar[1]-axi,2)/(pow(1-axi,3)+pow(axi,3))*2*intVar[1]; // xi2
    //   //
    //   //
    //   // integration in phi'
    //   //
    //   double cosphipmax = (xi2+xi-xi0)/2./sqrt(xi*xi2);
    //   //if(cosphipmax<1) cout << "Cosphipmax: " << cosphipmax << "  " << t2 << endl;
    //   double phipmin = (cosphipmax>1? 0 : myacos(cosphipmax));
    //   double phip = phipmin + 2*(M_PI-phipmin)*intVar[4];   jac*=2*(M_PI-phipmin);
    //   //phip = phipmin + 2*(M_PI-phipmin)*0.99;   jac*=2*(M_PI-phipmin);///////////////
    //   //cout << "prim: " << phip << endl;
    //   xi1 = xi+xi2-2*sqrt(xi*xi2)*cos(phip);
    //   //cout << intVar[1] << "  " << t2 << endl;
    //   if(xi2!=xi2) cout << yxi << "  " << axi << "  " << xi << "  " << xi2 << "  " << intVar[1] << "  " << phipmin << "  " << xi1 << "  " << xi0 << endl;
    //   //
    //   x1  = sqrt(x/z)*exp( etabar);
    //   x2  = sqrt(x/z)*exp(-etabar);
    //   return fabs(jac) * QQbar->diffXS_Y_pt_v2(xi, xi2, z, eta, phip, xi1) /z;
    // }
    double intQQbarSingleQuarkYq_v2(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 5);
      double jac = 1;
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double xiF = muFrat*muFrat;
      double pthat = QQbar->GetPthat();
      //
      // y,   x = partonic on-shell, arguments of C
      // eta, z = partonic off-shell, arguments of {\cal C}
      // etabar = y-eta
      // x/z         <= exp(-2|etabar|)   <<------ arguments of U <= 1
      // z(1+pthat2) <= exp(-2|eta|)      <<------ z1,z2 <= 1
      // prodotto:
      // A^2 <= exp(-2|eta|-2|etabar|) <= exp(-2|eta+etabar|) = exp(-2|y|)
      // from the first inequality we get:  A e^y <= exp(2etabar) <= e^y/A
      // from the condition 1/sqrt(z) >= B+sqrt(B^2-pthat^2) (see paper for the notation), together with x/z <= exp(-2|etabar|),
      // we find that B+sqrt(B^2-pthat^2) <= exp(-|etabar|)/sqrt(x), out of which we get a more stringent condition:
      // (A*exp(y)-x*pthat^2)/(1-A*exp(-y)) <= exp(2etabar) <= (1-A*exp(y))/(A*exp(-y)-x*pthat^2)
      //
      //
      // integration in etabar
      //
      double y = rapY;  // this is y, the argument of the on-shell coefficient
      double A = sqrt(x*(1+pthat*pthat));
      if(A>exp(-fabs(y))) cout << "Error: x, y and pt are incompatible" << endl;
      double exp2etabarmax = (1-A*exp(y))/(A*exp(-y)-x*pthat*pthat);
      double exp2etabarmin = (A*exp(y)-x*pthat*pthat)/(1-A*exp(-y));
      //cout << exp2etabarmax << "  " << exp2etabarmin << endl;
      if(exp2etabarmax < exp2etabarmin) cout << "Problem: etabar limits inverted" << endl;
      double etabar = 0.5*log(exp2etabarmin) + 0.5*(log(exp2etabarmax)-log(exp2etabarmin))*intVar[2]; // this is etabar, the integration variable
      //etabar = -1.69;
      //cout << etabar << endl;
      jac *= 0.5*(log(exp2etabarmax)-log(exp2etabarmin));
      //
      //
      // integration in z
      //
      double eta = y - etabar;  // this is eta = y-etabar, the argument of the off-shell coefficient
      double B = sqrt(1+pthat*pthat) * cosh(eta);
      double zmin = x*exp(2*fabs(etabar));  // comes from  x/z <= exp(-2|etabar|)
      double zmax = 1./pow(B+sqrt(B*B-pthat*pthat),2);
      if(zmax < zmin) {
	if(zmin-zmax>1e-10) cout << "Problem: z limits inverted: " << zmin << " < " << zmax << "   difference: " << zmin-zmax << endl;
	else zmin = zmax;
      }
      double z = zmin+(zmax-zmin)*(1+cos(M_PI*intVar[3]))/2;  // this is z, argument of {\cal C}
      //cout << "z = " << z << endl;
      jac *= (zmax-zmin)*M_PI*sin(M_PI*intVar[3])/2;
      //
      //
      // integration in xi, xi2
      //
      double stuff = pthat*pthat + 1/z - 2*B/sqrt(z);
      double sqrtstuff = sqrt(stuff);
      if(stuff<0 && stuff>-1e-10) sqrtstuff = 0;
      if(sqrtstuff != sqrtstuff) cout << "NaN again... " << stuff << endl;
      //cout << pthat*pthat + 1/z - sqrt(1+pthat*pthat)*(exp(eta)+exp(-eta))/sqrt(z) << endl;
      double sqrtximax = sqrtstuff + pthat; //min( sqrtstuff + pthat, sqrt(1./z-4));
      double sqrtximin = sqrtstuff - pthat; // it's the abs value of this, but then it is squared...
      if(sqrtximax < fabs(sqrtximin)) cout << "Problem: xi limits inverted" << endl;
      double xi0=xiF*exp(-1./asb0); double mult=2;//0.2;
      double ximin=sqrtximin*sqrtximin;  double ximax=sqrtximax*sqrtximax;
      double x0 = intVar[0], epsilon = 1e-5;
      if(intVar[0]<  epsilon) x0 =   epsilon;
      if(intVar[0]>1-epsilon) x0 = 1-epsilon;
      //double u1=intVar[0]*log(t1max/t1min);
      //double t1=t1min*exp(u1);     jac*=t1*log(t1max/t1min);  // xi
      //double t1=t1min + (t1max-t1min)*x0;     jac*=(t1max-t1min);  // xi
      double xi = ximin + (ximax-ximin)*(1+cos(M_PI*x0))/2;     jac*=(ximax-ximin)*M_PI*sin(M_PI*x0)/2;  // xi
      //if(t1 != t1) cout << "NaN ...  " << t1max << "  " << t1min << "  " << t1 << endl;
      double xi2min=xi0;  double xi2max=xiF*pow(xiF/xi0,mult);
      //double t2=t2min + (t2max-t2min)*intVar[1];     jac*=(t2max-t2min); // xi2
      //double u2=intVar[1]*log(t2max/t2min);     double t2=t2min*exp(u2);     jac*=t2*log(t2max/t2min); // xi2
      double yxi=max(1e-10,log(xi/xi2min)/log(xi2max/xi2min));  double yxi2=pow(yxi*pow(1-yxi,2),1./3.);  double axi=yxi-yxi*(1-yxi)/yxi2+yxi2;  double y2 = (pow(intVar[1]*intVar[1]-axi,3)+pow(axi,3))/(pow(1-axi,3)+pow(axi,3));
      double u2=y2*log(xi2max/xi2min);     xi2=xi2min*exp(u2);     jac*=xi2*log(xi2max/xi2min)*3*pow(intVar[1]*intVar[1]-axi,2)/(pow(1-axi,3)+pow(axi,3))*2*intVar[1]; // xi2
      //
      //
      // integration in phi'
      //
      double cosphipmax = (xi2+xi-xi0)/2./sqrt(xi*xi2);
      //if(cosphipmax<1) cout << "Cosphipmax: " << cosphipmax << "  " << t2 << endl;
      double phipmin = (cosphipmax>1? 0 : myacos(cosphipmax));
      double phip = phipmin + 2*(M_PI-phipmin)*intVar[4];   jac*=2*(M_PI-phipmin);
      //phip = phipmin + 2*(M_PI-phipmin)*0.99;   jac*=2*(M_PI-phipmin);///////////////
      //cout << "prim: " << phip << endl;
      xi1 = xi+xi2-2*sqrt(xi*xi2)*cos(phip);
      //cout << intVar[1] << "  " << t2 << endl;
      if(xi2!=xi2) cout << yxi << "  " << axi << "  " << xi << "  " << xi2 << "  " << intVar[1] << "  " << phipmin << "  " << xi1 << "  " << xi0 << endl;
      //
      x1  = sqrt(x/z)*exp( etabar);
      x2  = sqrt(x/z)*exp(-etabar);
      return fabs(jac) * QQbar->diffXS_Y_pt_v2(xi, xi2, z, eta, phip, xi1) /z;
    }


    double intQQbarSingleQuarkYq_v3(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 5);
      double jac = 1;
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double xiF = muFrat*muFrat;
      double pthat = QQbar->GetPthat();
      //
      // y,   x = partonic on-shell, arguments of C
      // eta, z = partonic off-shell, arguments of {\cal C}
      // etabar = y-eta
      // x/z         <= exp(-2|etabar|)   <<------ arguments of U <= 1
      // z(1+pthat2) <= exp(-2|eta|)      <<------ z1,z2 <= 1
      // prodotto:
      // A^2 <= exp(-2|eta|-2|etabar|) <= exp(-2|eta+etabar|) = exp(-2|y|)
      // from the first inequality we get:  A e^y <= exp(2etabar) <= e^y/A
      // from the condition 1/sqrt(z) >= B+sqrt(B^2-pthat^2) (see paper for the notation), together with x/z <= exp(-2|etabar|),
      // we find that B+sqrt(B^2-pthat^2) <= exp(-|etabar|)/sqrt(x), out of which we get a more stringent condition:
      // (A*exp(y)-x*pthat^2)/(1-A*exp(-y)) <= exp(2etabar) <= (1-A*exp(y))/(A*exp(-y)-x*pthat^2)
      //
      //
      // integration in etabar
      //
      double y = rapY;  // this is y, the argument of the on-shell coefficient
      double A = sqrt(x*(1+pthat*pthat));
      if(A>exp(-fabs(y))) cout << "Error: x, y and pt are incompatible" << endl;
      double exp2etabarmax = (1-A*exp(y))/(A*exp(-y)-x*pthat*pthat);
      double exp2etabarmin = (A*exp(y)-x*pthat*pthat)/(1-A*exp(-y));
      //cout << exp2etabarmax << "  " << exp2etabarmin << endl;
      if(exp2etabarmax < exp2etabarmin) cout << "Problem: etabar limits inverted" << endl;
      double etabar = 0.5*log(exp2etabarmin) + 0.5*(log(exp2etabarmax)-log(exp2etabarmin))*intVar[2]; // this is etabar, the integration variable
      //etabar = -1.69;
      //cout << etabar << endl;
      jac *= 0.5*(log(exp2etabarmax)-log(exp2etabarmin));
      //
      //
      // integration in z
      //
      double eta = y - etabar;  // this is eta = y-etabar, the argument of the off-shell coefficient
      double B = sqrt(1+pthat*pthat) * cosh(eta);
      double zmin = x*exp(2*fabs(etabar));  // comes from  x/z <= exp(-2|etabar|)
      double zmax = 1./pow(B+sqrt(B*B-pthat*pthat),2);
      if(zmax < zmin) {
	if(zmin-zmax>1e-10) cout << "Problem: z limits inverted: " << zmin << " < " << zmax << "   difference: " << zmin-zmax << endl;
	else zmin = zmax;
      }
      double z = zmin+(zmax-zmin)*(1+cos(M_PI*intVar[3]))/2;  // this is z, argument of {\cal C}
      //cout << "z = " << z << endl;
      jac *= (zmax-zmin)*M_PI*sin(M_PI*intVar[3])/2;
      //
      //
      // integration in xi1, xi2
      //
      double stuff = pthat*pthat + 1/z - 2*B/sqrt(z);
      double sqrtstuff = sqrt(stuff);
      if(stuff<0 && stuff>-1e-10) sqrtstuff = 0;
      if(sqrtstuff != sqrtstuff) cout << "NaN again... " << stuff << endl;
      //cout << pthat*pthat + 1/z - sqrt(1+pthat*pthat)*(exp(eta)+exp(-eta))/sqrt(z) << endl;
      double sqrtximax = sqrtstuff + pthat; //min( sqrtstuff + pthat, sqrt(1./z-4));
      double sqrtximin = sqrtstuff - pthat;
      if(sqrtximin < 0) { cout << "sqrtxi_min = " << sqrtximin << " < 0" << endl;   sqrtximin = 0; }
      if(sqrtximax < sqrtximin) cout << "Problem: xi limits inverted" << endl;
      double xi0 = xiF*exp(-1./asb0); double mult=1;//0.2;
      //double xi1min = max(sqrtximin*sqrtximin/4, xi0);  double xi1max = pow(xiF/xi0,mult);
      double xi1min = xi0;  double xi1max = pow(xiF/xi0,mult);
      int n = 5;  double u = pow(intVar[0],n);  jac *= n*pow(intVar[0],n-1);
      xi1 = xi1min + (xi1max-xi1min)*u;     jac*=(xi1max-xi1min);  // xi1
      //
      //double xi2min=xi1;  double xi2max=pow(sqrtximax+sqrt(xi1),2);
      double xi2min=max(xi1,pow(sqrtximin-sqrt(xi1),2));  double xi2max=pow(sqrtximax+sqrt(xi1),2);
      xi2 = xi2min + (xi2max-xi2min)*intVar[1];     jac*=(xi2max-xi2min);  // xi2
      //
      //
      // integration in phi
      //
      double cosphimax = (sqrtximax*sqrtximax-xi1-xi2)/2./sqrt(xi1*xi2);
      double cosphimin = (sqrtximin*sqrtximin-xi1-xi2)/2./sqrt(xi1*xi2);
      if(cosphimax> 1) cosphimax =  1; //cout << "Cosphimax-1: " << cosphimax-1 << endl;
      if(cosphimin<-1) cosphimin = -1; //cout << "Cosphimin+1: " << cosphimin+1 << endl;
      if(cosphimax<cosphimin) cout << "What? Cosphimax=" << cosphimax << " < cosphimin=" << cosphimin << endl;
      double phimin = myacos(cosphimax);
      double phimax = myacos(cosphimin);
      double phi = phimin + (phimax-phimin)*intVar[4];   jac*=phimax-phimin;
      //if(xi2!=xi2) cout << yxi << "  " << axi << "  " << xi << "  " << xi2 << "  " << intVar[1] << "  " << phipmin << "  " << xi1 << "  " << xi0 << endl;
      //
      x1  = sqrt(x/z)*exp( etabar);
      x2  = sqrt(x/z)*exp(-etabar);
      return fabs(jac)/z * ( QQbar->diffXS_Y_pt_v3(xi1, xi2, z, eta,  phi) +
			     QQbar->diffXS_Y_pt_v3(xi1, xi2, z, eta, -phi) +
			     QQbar->diffXS_Y_pt_v3(xi2, xi1, z, eta,  phi) +
			     QQbar->diffXS_Y_pt_v3(xi2, xi1, z, eta, -phi) +
			     0 );
    }
    double intQQbarSingleQuarkYqAux(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 2);
      double jac = 1;
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double xiF = muFrat*muFrat;
      double pthat = QQbar->GetPthat();
      //
      // y,   x = partonic on-shell, arguments of C
      // eta, z = partonic off-shell, arguments of {\cal C}
      // etabar = y-eta = 1/2 log(x/z)
      // x/z         <= 1   <<------ arguments of U <= 1
      // z(1+pthat2) <= exp(-2|eta|)      <<------ z1,z2 <= 1
      // prodotto:
      // A^2 <= exp(-2|eta|)
      // from the condition 1/sqrt(z) >= B+sqrt(B^2-pthat^2) (see paper for the notation), together with x/z <= 1, we get the condition:
      // x <= z <= x(1-A*exp(y))/(A*exp(-y)-x*pthat^2)
      //
      //
      // integration in etabar
      //
      double y = rapY;  // this is y, the argument of the on-shell coefficient
      //cout << y << endl;
      if((-x*pthat*pthat+exp(-y)*sqrt(x*(1+pthat*pthat)))<0) cout << "AAAAAAAAA" << endl;  // what are these?
      if((-x*pthat*pthat+exp( y)*sqrt(x*(1+pthat*pthat)))<0) cout << "BBBBBBBBB" << endl;
      double A = sqrt(x*(1+pthat*pthat));
      //
      //
      // integration in z
      //
      double zmin = x;  // comes from  x/z <= 1
      //double zmax = x*(1-A*exp(y))/(A*exp(-y));  // comes from X>=2B
      double zmax = x*(1-A*exp(y))/(A*exp(-y)-x*pthat*pthat);
      //double xi0 = xiF*exp(-1./asb0); double mult=1;//0.2;
      //zmax = (1-A*exp(y))/(A/x/exp(y)+xi0+2*sqrt(xi0)*pthat)*0.8;
      if(zmax < zmin) {
	if(zmin-zmax>1e-10) cout << "Problem: z limits inverted: " << zmin << " < " << zmax << "   difference: " << zmin-zmax << endl;
	else zmin = zmax;
      }
      int n1 = 2;
      double z = zmin+(zmax-zmin)*(1+cos(M_PI*pow(intVar[1],n1)))/2;  // this is z, argument of {\cal C}
      jac *= (zmax-zmin)*M_PI*sin(M_PI*pow(intVar[1],n1))/2 * n1*pow(intVar[1],n1-1);
      //
      //
      // integration in xi
      //
      double eta = y - 0.5*log(z/x);  // this is eta, the argument of the off-shell coefficient
      double B = sqrt(1+pthat*pthat) * cosh(eta);
      double stuff = pthat*pthat + 1/z - 2*B/sqrt(z);
      double sqrtstuff = sqrt(stuff);
      if(stuff<0 && stuff>-1e-10) sqrtstuff = 0;
      if(sqrtstuff != sqrtstuff) cout << "sqrt(stuff) is NaN again... " << stuff << endl;
      //cout << pthat*pthat + 1/z - sqrt(1+pthat*pthat)*(exp(eta)+exp(-eta))/sqrt(z) << endl;
      double sqrtximax = sqrtstuff + pthat; //min( sqrtstuff + pthat, sqrt(1./z-4));
      double sqrtximin = sqrtstuff - pthat;
      //if(sqrtximin < 0) { cout << "sqrtxi_min = " << sqrtximin << " < 0" << endl;   sqrtximin = 0; }
      if(sqrtximax < fabs(sqrtximin)) cout << "Problem: xi limits inverted" << endl;
      double xi0 = xiF*exp(-1./asb0); double mult=1;//0.2;
      double xi1min = max(sqrtximin*sqrtximin, xi0);  double xi1max = sqrtximax*sqrtximax;
      //double xi1min = sqrtximin*sqrtximin;  double xi1max = sqrtximax*sqrtximax;
      if(xi1max < xi1min) cout << "Warning: xi1max=" << xi1max << " < xi1min=" << xi1min << endl;
      int n = 1;
      //double u = pow(intVar[0],n);  jac *= n*pow(intVar[0],n-1);
      //xi1 = xi1min + (xi1max-xi1min)*u;     jac*=(xi1max-xi1min);  // xi1
      xi1 = xi1min+(xi1max-xi1min)*(1+cos(M_PI*pow(intVar[0],n)))/2;
      jac *= (xi1max-xi1min)*M_PI*sin(M_PI*pow(intVar[0],n))/2 * n*pow(intVar[0],n-1);
      //
      //
      x1  = x/z;
      x2  = x/z;
      return fabs(jac)/z * QQbar->diffXS_Y_pt_Aux_v1(xi1, z, eta);
    }
    double intQQbarSingleQuarkYqAux_v1(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 2);
      double jac = 1;
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double xiF = muFrat*muFrat;
      double pthat = QQbar->GetPthat();
      //
      // y,   x = partonic on-shell, arguments of C
      // eta, z = partonic off-shell, arguments of {\cal C}
      // etabar = y-eta = 1/2 log(x/z)
      // x/z         <= 1   <<------ arguments of U <= 1
      // z(1+pthat2) <= exp(-2|eta|)      <<------ z1,z2 <= 1
      // prodotto:
      // A^2 <= exp(-2|eta|)
      // from the condition 1/sqrt(z) >= B+sqrt(B^2-pthat^2) (see paper for the notation), together with x/z <= 1, we get the condition:
      // x <= z <= x(1-A*exp(y))/(A*exp(-y)-x*pthat^2)
      //
      //
      // integration in etabar
      //
      double y = rapY;  // this is y, the argument of the on-shell coefficient
      //cout << y << endl;
      if((-x*pthat*pthat+exp(-y)*sqrt(x*(1+pthat*pthat)))<0) cout << "AAAAAAAAA" << endl;  // what are these?
      if((-x*pthat*pthat+exp( y)*sqrt(x*(1+pthat*pthat)))<0) cout << "BBBBBBBBB" << endl;
      double A = sqrt(x*(1+pthat*pthat));
      //
      //
      // integration in z
      //
      double zmin = x;  // comes from  x/z <= 1
      double zmax = x*(1-A*exp(y))/(A*exp(-y)-x*pthat*pthat);
      //int n1 = 1;
      //double z = zmin+(zmax-zmin)*pow(intVar[1],n1);  // this is z, argument of {\cal C}
      //jac *= (zmax-zmin)* n1*pow(intVar[1],n1-1);
      int n1 = 2;
      double z = zmin+(zmax-zmin)*(1+cos(M_PI*pow(intVar[1],n1)))/2;  // this is z, argument of {\cal C}
      jac *= (zmax-zmin)*M_PI*sin(M_PI*pow(intVar[1],n1))/2 * n1*pow(intVar[1],n1-1);
      //
      //
      // integration in xi
      //
      double eta = y - 0.5*log(z/x);  // this is eta, the argument of the off-shell coefficient
      double B = sqrt(1+pthat*pthat) * cosh(eta);
      double stuff = pthat*pthat + 1/z - 2*B/sqrt(z);
      double sqrtstuff = sqrt(stuff);
      if(stuff<0 && stuff>-1e-10) sqrtstuff = 0;
      if(sqrtstuff != sqrtstuff) cout << "sqrt(stuff) is NaN again... " << stuff << endl;
      //cout << pthat*pthat + 1/z - sqrt(1+pthat*pthat)*(exp(eta)+exp(-eta))/sqrt(z) << endl;
      double sqrtximax = sqrtstuff + pthat; //min( sqrtstuff + pthat, sqrt(1./z-4));
      double sqrtximin = sqrtstuff - pthat;
      //if(sqrtximin < 0) { cout << "sqrtxi_min = " << sqrtximin << " < 0" << endl;   sqrtximin = 0; }
      if(sqrtximax < fabs(sqrtximin)) cout << "Problem: xi limits inverted" << endl;
      double xi0 = xiF*exp(-1./asb0); double mult=1;//0.2;
      double xi1min = max(sqrtximin*sqrtximin, xi0);  double xi1max = sqrtximax*sqrtximax;
      //double xi1min = sqrtximin*sqrtximin;  double xi1max = sqrtximax*sqrtximax;
      if(xi1max < xi1min) cout << "Warning: xi1max=" << xi1max << " < xi1min=" << xi1min << endl;
      int n = 1;
      //double u = pow(intVar[0],n);  jac *= n*pow(intVar[0],n-1);
      double u = (1+cos(M_PI*pow(intVar[0],n)))/2;  jac *= M_PI*sin(M_PI*pow(intVar[0],n))/2 * n*pow(intVar[0],n-1);
      xi1 = xi1min + (xi1max-xi1min)*u;     jac*=(xi1max-xi1min);  // xi1
      //
      // double xi0 = xiF*exp(-1./asb0); double mult=0.5;//0.2;
      // double xi1min = xi0;  double xi1max = pow(xiF/xi0,mult);
      // if(xi1max < xi1min) cout << "Warning: xi1max=" << xi1max << " < xi1min=" << xi1min << endl;
      // int n = 4;
      // double u = pow(intVar[0],n);  jac *= n*pow(intVar[0],n-1);
      // //double u = 1+cos(M_PI*pow(intVar[0],n)))/2;  jac *= M_PI*sin(M_PI*intVar[0])/2 * n*pow(intVar[0],n-1);
      // xi1 = xi1min + (xi1max-xi1min)*u;     jac*=(xi1max-xi1min);  // xi1
      //
      x1  = x/z;
      x2  = x/z;
      //double eta = y - 0.5*log(z/x);  // this is eta, the argument of the off-shell coefficient
      return fabs(jac)/z * QQbar->diffXS_Y_pt_Aux_v1(xi1, z, eta);
    }
    double intQQbarSingleQuarkYqAux_v2(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 2);
      double jac = 1;
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double xiF = muFrat*muFrat;
      double pthat = QQbar->GetPthat();
      //
      // y,   x = partonic on-shell, arguments of C
      // eta, z = partonic off-shell, arguments of {\cal C}
      //
      //
      // integration in etabar
      //
      double y = rapY;  // this is y, the argument of the on-shell coefficient
      double A = sqrt(x*(1+pthat*pthat));
      if((1+x*pthat*pthat)/2/A/cosh(y)-1<-1e-10) cout << "Nooooooo" << endl;
      //
      //
      // integration in theta'
      //
      double thetap = 2*M_PI*intVar[1];
      jac *= 2*M_PI;
      //
      //
      // integration in xi
      //
      double xi0 = xiF*exp(-1./asb0); double mult=1;//0.2;
      double xi1min = xi0;  double xi1max = pow(xiF/xi0,mult);
      if(xi1max < xi1min) cout << "Warning: xi1max=" << xi1max << " < xi1min=" << xi1min << endl;
      int n = 4;
      double u = pow(intVar[0],n);  jac *= n*pow(intVar[0],n-1);
      //double u = 1+cos(M_PI*pow(intVar[0],n)))/2;  jac *= M_PI*sin(M_PI*intVar[0])/2 * n*pow(intVar[0],n-1);
      xi1 = xi1min + (xi1max-xi1min)*u;     jac*=(xi1max-xi1min);  // xi1
      //
      //
      double num = 1-A*exp(y);
      double den = A/x*exp(-y)+xi1-2*pthat*sqrt(xi1)*cos(thetap);
      double z = num/den;
      double eta = y - 0.5*log(z/x);  // this is eta, the argument of the off-shell coefficient
      x1  = x/z;
      x2  = x/z;
      //
      //if(z<x) return 0;
      double zmax = min(1., x*num/(A*exp(-y)-x*pthat*pthat));           // min should always select the second argument
      double zmin = max(x,  x*(A*exp(y)-x*pthat*pthat)/(1-A*exp(-y)));  // max should always select the first  argument
      //if(z>zmax) { cout << "max" << endl;  return 0; }
      //if(z<zmin) { cout << "min" << endl;  return 0; }
      if(z>zmax) return 0;  // this should never happen
      if(z<zmin) return 0;  // this instead may happen, for small values of xi and perhaps for some combinations of xi, pt and theta'
      return fabs(jac) / fabs(num) * QQbar->diffXS_Y_pt_Aux_v2(xi1, z, eta);
    }
    double intQQbarSingleQuarkYqAux_v2_subplus(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 2);
      double jac = 1;
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double xiF = muFrat*muFrat;
      double pthat = QQbar->GetPthat();
      //
      // y,   x = partonic on-shell, arguments of C
      // eta, z = partonic off-shell, arguments of {\cal C}
      //
      //
      // integration in etabar
      //
      double y = rapY;  // this is y, the argument of the on-shell coefficient
      double A = sqrt(x*(1+pthat*pthat));
      //if(1+x*pthat*pthat<2*A*cosh(y)) cout << "Nooooooo" << endl;
      if((1+x*pthat*pthat)/2/A/cosh(y)-1<-1e-10) cout << "Nooooooo" << endl;
      //
      //
      // integration in theta'
      //
      double thetap = 2*M_PI*intVar[1];
      jac *= 2*M_PI;
      //
      //
      // integration in xi
      //
      double xi0 = xiF*exp(-1./asb0); double mult=1;//0.2;
      double xi1min = xi0;  double xi1max = pow(xiF/xi0,mult);
      if(xi1max < xi1min) cout << "Warning: xi1max=" << xi1max << " < xi1min=" << xi1min << endl;
      int n = 4;
      double u = pow(intVar[0],n);  jac *= n*pow(intVar[0],n-1);
      //double u = 1+cos(M_PI*pow(intVar[0],n)))/2;  jac *= M_PI*sin(M_PI*intVar[0])/2 * n*pow(intVar[0],n-1);
      xi1 = xi1min + (xi1max-xi1min)*u;     jac*=(xi1max-xi1min);  // xi1
      //
      //
      double num = 1-A*exp(y);
      double den = A/x*exp(-y);
      double z = num/den;
      double eta = y - 0.5*log(z/x);  // this is eta, the argument of the off-shell coefficient
      x1  = x/z;
      x2  = x/z;
      //
      //if(z<x) return 0;
      double zmax = min(1., x*num/(A*exp(-y)-x*pthat*pthat));           // min should always select the second argument
      double zmin = max(x,  x*(A*exp(y)-x*pthat*pthat)/(1-A*exp(-y)));  // max should always select the first  argument
      //if(z>zmax) { cout << "max" << endl;  return 0; }
      //if(z<zmin) { cout << "min" << endl;  return 0; }
      if(z>zmax) return 0;  // this should never happen
      if(z<zmin) return 0;  // this instead may happen, for large values of y close to the boundary, especially if x is not small
      if(xi1 > 1) return 0;  // it is better to have this theta function so that the integration range is the same of the other
      return fabs(jac) / fabs(num) * QQbar->diffXS_Y_pt_Aux_v2_subplus(z, eta);
    }


    double intCheck0(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par) {
      checkDimension(dim, 1);
      xi1 = intVar[0];
      return 1;
    }
    double intCheck1(int dim, double *intVar, double &xi1, double &xi2, double &x1, double &x2, double x, double rapY, double muFrat, double asb0, void *par, bool sub) {
      checkDimension(dim, 1);
      double xiF = muFrat*muFrat;
      double xi0 = xiF*exp(-1./asb0);
      double ximin = xi0;
      xi1 = ximin+(1-ximin)*intVar[0];
      double jac = 1-ximin;
      //
      double res = 1-xi1;
      if(sub) res -= 1;
      //
      return fabs(jac) * res;
    }




    bool checkU(double xi, double xi0) {
      if(xi<xi0) {
	//cout << "cazzo!  xi0=" << xi0 << "  xi-xi0=" << xi-xi0 << "  xi=" << xi << endl;
	return true;
	if(fabs(xi-xi0) < xi0*1e-3) {
	  xi = xi0;
	  return false;
	} else {
	  cout << "cazzo!  xi0=" << xi0 << "  xi-xi0=" << xi-xi0 << "  xi=" << xi << endl;
	  return true;
	}
      }
      return false;
    }


    //
    // integrand of the form:  U'(x, xi, muF) * Coff(xi, pars)   or   - U(x, xi, muF) * dCoff/dxi(xi, pars)
    //
    //double integrand(double xi, double x, xTableU *tU, process p, double asb0, double muFrat, void *par) { //, double *intVars) {
    double integrand(double x, double rapY, xTableU *tU, process p, double muFrat, void *par, double *intVars, int nf, double as) {
      double xiF = muFrat*muFrat;
      double asb0 = as*beta0(nf);
      double xi0 = exp(-1./asb0);
      //
      double xi1, xi2, x1, x2;
      double res = 0;
      //
      if(p == check0) {
	res = intCheck0(1, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	x1 = x;
	if(x1>1) return 0;
	double xiU1 = xi1/xiF;
	if(checkU(xiU1,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	//double dU1 = DapproxUreg(xiU1, x1, as, asb0/as, 0, as*3/M_PI, true);
	res *= dU1;
	//cout << res << endl;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << xi1 << "  " << xiU1 << "  " << dU1 << endl;
	  res = 0;
	}
      }
      if(p == check1) {
	//res = intCheck1(1, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par, false);
	res = intCheck1(1, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par, true);
	res *= as;
	x1 = x;
	if(x1>1) return 0;
	double xiU1 = xi1/xiF;
	if(checkU(xiU1,xi0)) return 0;
	double dU1, U1 = tU->evalU(x1, xiU1, dU1);
	//double dU1 = tU->evalDU(x1, xiU1);
	res *= dU1;
	//res = U1;
	//cout << res << endl;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << xi1 << "  " << xiU1 << "  " << dU1 << endl;
	  res = 0;
	}
      }
      if(p == QQbarPairM2YLO_qqbarchannel) {
	res = intQQbarPairM2YLO_qqbarchannel(1, intVars, par);
	if(res!=res) {
	  cout << res << "  " << intVars[0] << endl;
	  res = 0;
	}
      }
      return res;
    }

    //
    // integrand of the form:  U'(x1, xi1, muF) * U'(x2, xi2, muF) * Coff(xi1, xi2, pars)
    //
    double integrand2d(int dim, double x, double rapY, xTableU *tU, process p, double muFrat, void *par, double *intVars, int nf, double as) {
      double xiF = muFrat*muFrat;
      double asb0 = as*beta0(nf);
      double xi0 = exp(-1./asb0);
      //
      double xi1, xi2, x1, x2;
      double res = 0;
      //
      // if(p == check) {
      // 	res = intCheck(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
      // 	x1 = x;
      // 	if(x1>1) return 0;
      // 	double xiU1 = xi1/xiF;
      // 	if(checkU(xiU1,xi0)) return 0;
      // 	double dU1;  tU->evalU(x1, xiU1, dU1);
      // 	res = dU1;
      // 	if(res!=res) {
      // 	  cout << res << "  " << x1 << "  " << xi1 << "  " << xiU1 << "  " << dU1 << endl;
      // 	  res = 0;
      // 	}
      // }


      // QQbar Pair, triple differential
      if(p == QQbarPairM2Yq) {
	res = intQQbarPairM2Yq(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1 || x2>1) return 0;
	double xiU1 = xi1/xiF;
	double xiU2 = xi2/xiF;
	if(checkU(xiU1,xi0)) return 0;
	if(checkU(xiU2,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	double dU2;  tU->evalU(x2, xiU2, dU2);
	double dU1a; tU->evalU(x2, xiU1, dU1a); // swapped x1 and x2
	double dU2a; tU->evalU(x1, xiU2, dU2a);
	res *= dU1*dU2 + dU1a*dU2a;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << x2 << "  " << xi1 << "  " << xi2 << "  " << xiU1 << "  " << xiU2 << "  " << intVars[0] << "  " << intVars[1] << "  " << intVars[2] << "  " << intVars[3] << "  " << dU1 << "  " << dU2 << endl;
	  res = 0;
	}
      }
      /*
      if(p == QQbarPairM2Yq_v2) {
	res = intQQbarPairM2Yq_v2(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1 || x2>1) return 0;
	double xiU1 = xi1/xiF;
	double xiU2 = xi2/xiF;
	if(checkU(xiU1,xi0)) return 0;
	if(checkU(xiU2,xi0)) return 0;
	double dU1; tU->evalU(x1, xiU1, dU1);
	double dU2; tU->evalU(x2, xiU2, dU2);
	//double cosphipmax = (xi2+qthat*qthat-xi0)/2./qthat/sqrt(xi2);
	//double phipmin = (cosphipmax>1? 0 : acos(cosphipmax));
	//double phip = phipmin + 2*(M_PI-phipmin)*intVars[1];
	res *= dU1*dU2;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << x2 << "  " << xi1 << "  " << xi2 << "  " << xiU1 << "  " << xiU2 << "  " << intVars[0] << "  " << intVars[1] << "  " << intVars[2] << "  " << intVars[3] << "  " << dU1 << "  " << dU2 << endl;
	  res = 0;
	}
      }
      */
      if(p == QQbarPairM2Yq_part1) {
	double sub = 0, xi2s = 0;
	double integr = intQQbarPairM2Yq_part1(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par, xi2s, sub);
	if(x1>1 || x2>1) return 0;
	double xiU1  = xi1/xiF;
	double xiU2  = xi2/xiF;
	double xiU2s = xi2s/xiF;
	if(checkU(xiU1,xi0)) return 0;
	if(checkU(xiU2,xi0)) return 0;
	if(checkU(xiU2s,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	double dU2;  tU->evalU(x2, xiU2, dU2);
	double dU2s; tU->evalU(x2, xiU2s, dU2s);
	double dU1a; tU->evalU(x2, xiU1, dU1a); // swapped x1 and x2
	double dU2a; tU->evalU(x1, xiU2, dU2a);
	double dU2sa; tU->evalU(x1, xiU2s, dU2sa);
	res = dU1*(integr*dU2-sub*dU2s) + dU1a*(integr*dU2a-sub*dU2sa);
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << x2 << "  " << xi1 << "  " << xi2 << "  " << xiU1 << "  " << xiU2 << "  " << intVars[0] << "  " << intVars[1] << "  " << intVars[2] << "  " << intVars[3] << "  " << dU1 << "  " << dU2 << endl;
	  res = 0;
	}
      }
      if(p == QQbarPairM2Yq_part2) {
	res = intQQbarPairM2Yq_part2(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1 || x2>1) return 0;
	double xiU1 = xi1/xiF;
	double xiU2 = xi2/xiF;
	if(checkU(xiU1,xi0)) return 0;
	if(checkU(xiU2,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	double dU2;  tU->evalU(x2, xiU2, dU2);
	double dU1a; tU->evalU(x2, xiU1, dU1a); // swapped x1 and x2
	double dU2a; tU->evalU(x1, xiU2, dU2a);
	res *= dU1*dU2 + dU1a*dU2a;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << x2 << "  " << xi1 << "  " << xi2 << "  " << xiU1 << "  " << xiU2 << "  " << intVars[0] << "  " << intVars[1] << "  " << intVars[2] << "  " << intVars[3] << "  " << dU1 << "  " << dU2 << endl;
	  res = 0;
	}
      }
      if(p == QQbarPairM2YqAux) {
	res = intQQbarPairM2Yq_Aux(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1) return 0;
	double xiU1 = xi1/xiF;
	if(checkU(xiU1,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	res *= dU1;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << xi1 << "  " << xiU1 << "  " << intVars[0] << "  " << intVars[1]  << "  " << dU1 << endl;
	  res = 0;
	}
      }


      // QQbar Pair, double differential
      if(p == QQbarPairM2Y_v2) {
	res = intQQbarPairM2Y_v2(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1 || x2>1) return 0;
	double xiU1 = xi1/xiF;
	double xiU2 = xi2/xiF;
	if(checkU(xiU1,xi0)) return 0;
	if(checkU(xiU2,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	double dU2;  tU->evalU(x2, xiU2, dU2);
	double dU1a; tU->evalU(x2, xiU1, dU1a); // swapped x1 and x2
	double dU2a; tU->evalU(x1, xiU2, dU2a);
	res *= dU1*dU2 + dU1a*dU2a;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << x2 << "  " << xi1 << "  " << xi2 << "  " << intVars[0] << "  " << intVars[1] << "  " << intVars[2] << "  " << intVars[3] << "  " << dU1 << "  " << dU2 << endl;
	  res = 0;
	}
      }
      if(p == QQbarPairM2Y) {
	res = intQQbarPairM2Y(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1 || x2>1) return 0;
	double xiU1 = xi1/xiF;
	double xiU2 = xi2/xiF;
	if(checkU(xiU1,xi0)) return 0;
	if(checkU(xiU2,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	double dU2;  tU->evalU(x2, xiU2, dU2);
	res *= dU1*dU2;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << x2 << "  " << xi1 << "  " << xi2 << "  " << intVars[0] << "  " << intVars[1] << "  " << intVars[2] << "  " << intVars[3] << "  " << intVars[4] << "  " << dU1 << "  " << dU2 << endl;
	  res = 0;
	}
      }
      if(p == QQbarPairM2Y_part1) {
	res = intQQbarPairM2Y_part1(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1 || x2>1) return 0;
	double xiU1 = xi1/xiF;
	double xiU2 = xi2/xiF;
	if(checkU(xiU1,xi0)) return 0;
	if(checkU(xiU2,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	double dU2;  tU->evalU(x2, xiU2, dU2);
	double dU1a; tU->evalU(x2, xiU1, dU1a); // swapped x1 and x2
	double dU2a; tU->evalU(x1, xiU2, dU2a);
	res *= dU1*dU2 + dU1a*dU2a;
	//res *= dU1*dU2;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << x2 << "  " << xi1 << "  " << xi2 << "  " << intVars[0] << "  " << intVars[1] << "  " << intVars[2] << "  " << intVars[3] << "  " << intVars[4] << "  " << dU1 << "  " << dU2 << endl;
	  res = 0;
	}
      }
      if(p == QQbarPairM2YAux) {
	res = intQQbarPairM2YAux(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1) return 0;
	double xiU1 = xi1/xiF;
	if(checkU(xiU1,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	res *= dU1;
	//cout << res << endl;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << xi1 << "  " << intVars[0] << "  " << intVars[1] << "  " << dU1 << endl;
	  res = 0;
	}
      }
      if(p == QQbarPairM2YAux_part2) {
	res = intQQbarPairM2YAux_part2(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1) return 0;
	double xiU1 = xi1/xiF;
	if(checkU(xiU1,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	//double dU1 = DapproxUreg(xiU1, x1, as, asb0/as, 0, as*3/M_PI, true);
	res *= dU1;
	//cout << res << endl;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << xi1 << "  " << intVars[0] << "  " << intVars[1] << "  " << dU1 << endl;
	  res = 0;
	}
      }
      if(p == QQbarPairM2YLO) {
	res = intQQbarPairM2YLO(dim, intVars, par);
	if(res!=res) {
	  cout << res << "  " << intVars[0] << "  " << intVars[1] << "  " << intVars[2] << endl;
	  res = 0;
	}
      }
      //
      //
      // SINGLE QUARK
      //
      if(p == QQbarSingleQuarkYq) {
	res = intQQbarSingleQuarkYq(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1 || x2>1) return 0;
	double xiU1 = xi1/xiF;
	double xiU2 = xi2/xiF;
	if(checkU(xiU1,xi0)) return 0;
	if(checkU(xiU2,xi0)) return 0;
	double dU1;  tU->evalU(x1, xiU1, dU1);
	double dU2;  tU->evalU(x2, xiU2, dU2);
	double dU1a; tU->evalU(x2, xiU1, dU1a); // swapped x1 and x2
	double dU2a; tU->evalU(x1, xiU2, dU2a);
	//res = dU1 * dU2 * QQbar->diffXS_Y_pt(xi1, xi2, z, y, intVars[4]) + dU1a * dU2a * QQbar->diffXS_Y_pt(xi2, xi1, z, y, intVars[4]);  // I was using this one...
	res *= dU1*dU2 + dU1a*dU2a;
      }
      if(p == QQbarSingleQuarkYq_v2) {
	res = intQQbarSingleQuarkYq_v2(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	//res = intQQbarSingleQuarkYq_v3(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1 || x2>1) return 0;
	double xiU1 = xi1/xiF;
	double xiU2 = xi2/xiF;
	if(checkU(xiU1,xi0)) return 0;
	if(checkU(xiU2,xi0)) return 0;
	double dU1; tU->evalU(x1, xiU1, dU1);
	double dU2; tU->evalU(x2, xiU2, dU2);
	//double cosphipmax = (xi2+xi-xi0)/2./sqrt(xi*xi2);
	//double phipmin = (cosphipmax>1? 0 : acos(cosphipmax));
	//double phip = phipmin + 2*(M_PI-phipmin)*intVars[1];
	//double xi1test = xi+xi2-2*sqrt(xi*xi2)*cos(phip);
	//cout << xi1test << "  " << xi1 << endl;
	//cout << "dopo: " << phip << endl;
	res *= dU1 * dU2;
	if(res != res) {
	  //cout << res << "  " << xi1 << "  " << phip << "  " << xi << "  " << xi2 << "  " << z << "  " << eta << "  " << dU1 << "  " << dU2 << "  " << phinonprime(xi,xi2,phip) << endl;
	  res = 0;
	}
      }
      if(p == QQbarSingleQuarkYqAux) {
	//res = intQQbarSingleQuarkYqAux(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	res = intQQbarSingleQuarkYqAux_v1(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1) return 0;
	double xiU1 = xi1/xiF;
	if(checkU(xiU1,xi0)) return 0;
	double dU1; tU->evalU(x1, xiU1, dU1);
	//double dU1 = DapproxUreg(xiU1, x1, as, asb0/as, 0, as*3/M_PI, true);
	res *= dU1;
	//res *= as*P0(x1,nf)/xi1;
	if(res != res) {
	  //cout << res << "  " << xi1 << "  " << phip << "  " << xi << "  " << xi2 << "  " << z << "  " << eta << "  " << dU1 << "  " << dU2 << "  " << phinonprime(xi,xi2,phip) << endl;
	  res = 0;
	}
      }
      if(p == QQbarSingleQuarkYqAux_v2) {
	res = intQQbarSingleQuarkYqAux_v2(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1) return 0;
	double xiU1 = xi1/xiF;
	if(checkU(xiU1,xi0)) return 0;
	double dU1; tU->evalU(x1, xiU1, dU1);
	//double dU1 = DapproxUreg(xiU1, x1, as, asb0/as, 0, as*3/M_PI, true);
	res *= dU1;
	//res *= as*P0(x1,nf)/xi1;
	if(res != res) {
	  //cout << res << "  " << xi1 << "  " << phip << "  " << xi << "  " << xi2 << "  " << z << "  " << eta << "  " << dU1 << "  " << dU2 << "  " << phinonprime(xi,xi2,phip) << endl;
	  res = 0;
	}
      }
      //
      //cout << res << endl;
      return res;
    }


    double integrand2d_SUB(int dim, double x, double rapY, process p, double muFrat, void *par, double *intVars, int nf, double as) {
      double xi1, xi2, x1=0, x2=0;
      double asb0 = as*beta0(nf);
      double res = 0;
      //
      if(p == check1) {
	res = intCheck1(1, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, 0, par, true);
	x1 = x;
	if(x1>1) return 0;
	double P0x1 = P0(x1,nf);
	res *= as*as* P0x1 /xi1;
	if(res != res) {
	  cout << res << "  " << xi1 << "  " << x1 << "  " << P0x1 << endl;
	  //res = 0;
	}
      }
      if(p == QQbarPairM2Yq_part1) {
	double sub = 0, xi2s = 0;
	double integr = intQQbarPairM2Yq_part1(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par, xi2s, sub, true);
	//res = intQQbarPairM2Yq_SUB1(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, par);
	if(x1>1 || x2>1) return 0;
	//double xP0x1 = x1*P0(x1,nf);
	//double xP0x2 = x2*P0(x2,nf);
	double P0x1 = P0(x1,nf);
	double P0x2 = P0(x2,nf);
	res = as*as* 2* P0x1 * P0x2 /xi1 * (integr/xi2-sub/xi2s);
	if(res != res) {
	  cout << res << "  " << xi1 << "  " << x1 << "  " << x2 << "  " << P0x1 << "  " << P0x2 << endl;
	  res = 0;
	}
      }
      if(p == QQbarPairM2Yq_part2) {
	res = intQQbarPairM2Yq_part2(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par, true);
	if(x1>1 || x2>1) return 0;
	//double xP0x1 = x1*P0(x1,nf);
	//double xP0x2 = x2*P0(x2,nf);
	double P0x1 = P0(x1,nf);
	double P0x2 = P0(x2,nf);
	res *= as*as* 2* P0x1 * P0x2 /xi1 /xi2;
	if(res != res) {
	  cout << res << "  " << xi1 << "  " << x1 << "  " << x2 << "  " << P0x1 << "  " << P0x2 << endl;
	  res = 0;
	}
      }
      /*
      if(p == QQbarPairM2Yq_SUB4) {
	res = intQQbarPairM2Yq_SUB4(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, par);
	if(x1>1 || x2>1) return 0;
	//double xP0x1 = x1*P0(x1,nf);
	//double xP0x2 = x2*P0(x2,nf);
	double P0x1 = P0(x1,nf);
	double P0x2 = P0(x2,nf);
	res *= as*as* 2* P0x1 * P0x2 /xi1 /xi2;
	if(res != res) {
	  //res = 0;
	}
      }
      if(p == QQbarPairM2Yq_SUBc) {
	res = intQQbarPairM2Yq_SUBc(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, par);
	if(x1>1 || x2>1) return 0;
	//double xP0x1 = x1*P0(x1,nf);
	//double xP0x2 = x2*P0(x2,nf);
	double P0x1 = P0(x1,nf);
	double P0x2 = P0(x2,nf);
	res *= as*as* P0x1 * P0x2;
      }
      */
      if(p == QQbarPairM2YqAux) {
	res = intQQbarPairM2Yq_Aux(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1) return 0;
	double P0x1 = P0(x1,nf);
	res *= as* P0x1 /xi1;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << xi1 << "  " << intVars[0] << "  " << intVars[1] << endl;
	  res = 0;
	}
      }



      if(p == QQbarPairM2YAux) {
	res = intQQbarPairM2YAux(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1) return 0;
	double P0x1 = P0(x1,nf);
	res *= as* P0x1 /xi1;
	//cout << res << endl;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << xi1 << "  " << intVars[0] << "  " << intVars[1] << endl;
	  res = 0;
	}
      }
      if(p == QQbarPairM2YAux_part2) {
	res = intQQbarPairM2YAux_part2(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(x1>1) return 0;
	double P0x1 = P0(x1,nf);
	res *= as* P0x1 /xi1;
	//cout << res << endl;
	if(res!=res) {
	  cout << res << "  " << x1 << "  " << xi1 << "  " << intVars[0] << "  " << intVars[1] << endl;
	  res = 0;
	}
      }
      //
      //
      // SINGLE QUARK
      //
      if(p == QQbarSingleQuarkYqAux_v2) {
	res = intQQbarSingleQuarkYqAux_v2(dim, intVars, xi1, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	double P0x1 = ( x1>1 ? 0 : P0(x1,nf) );
	double xi1sub;
	double res_sub = intQQbarSingleQuarkYqAux_v2_subplus(dim, intVars, xi1sub, xi2, x1, x2, x, rapY, muFrat, asb0, par);
	if(fabs(xi1sub-xi1) > 1e-5) cout << "Warning: the two xi1's of the Ord(as) expansion of SQ Aux are inconsistent" << endl;
	double P0x1_sub = ( x1>1 ? 0 : P0(x1,nf) );
	res = as * (res*P0x1 - res_sub*P0x1_sub) /xi1;
	if(res != res) {
	  //cout << res << "  " << xi1 << "  " << phip << "  " << xi << "  " << xi2 << "  " << z << "  " << eta << "  " << dU1 << "  " << dU2 << "  " << phinonprime(xi,xi2,phip) << endl;
	  res = 0;
	}
      }
      //
      res *= pow(1-x1,2); // damping!!
      res *= pow(1-x2,2); // damping!!
      return res;
    }







    struct s_prdtv { process p; double d1, d2, d3, d4, d5; xTableU *tU; void *par; int nf; bool sub; };
    double integrandGSLx(double x, void *pars) {
      s_prdtv p = *(s_prdtv*)pars;
      double xx  = p.d1;
      double a0 = p.d2;  // as(Q^2)
      double muFrat = p.d4;
      double rapY = p.d5;
      //
      double *intVars = new double[1];
      intVars[0] = x;
      //
      double result = integrand(xx, rapY, p.tU, p.p, muFrat, p.par, intVars, p.nf, a0);
      //if(result==result) cout << result << endl;
      //
      //cout << setw(15) << xi << setw(15) << x[1] << setw(35) << res << endl;
      //delete tU;
      return result;
    }
    int integrandCUBAx(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
      //if(x[1]==0)
      s_prdtv p = *(s_prdtv*)pars;
      double xx  = p.d1;
      double a0 = p.d2;  // as(Q^2)
      //double b0 = p.d3;  // beta0
      double muFrat = p.d4;
      double rapY  = p.d5;
      //double xiF = muFrat*muFrat;
      //void *par = p.par;
      //process pr = p.p;
      //
      double *intVars = new double[*ndim];
      for(int i=0; i<*ndim; i++) intVars[i] = x[i];
      //
      if(!p.sub) res[0] = integrand2d    (*ndim, xx, rapY, p.tU, p.p, muFrat, p.par, intVars, p.nf, a0);
      else       res[0] = integrand2d_SUB(*ndim, xx, rapY,       p.p, muFrat, p.par, intVars, p.nf, a0);
      //
      if(res[0]!=res[0]) {
	cout << "Result is nan." << endl;
	res[0]=0;  // tappullo!!
      }
      return 0;
    }

    struct s_prdtvXX { s_prdtv pars; double *xx; int dim; };
    double test_integrandCUBAx(double x, void *p) {
      s_prdtvXX intparsXX = *(s_prdtvXX*) p;
      int dim = intparsXX.dim;
      const int ncomp = 1;
      double cres[ncomp];
      double *xx = new double[dim];
      for(int i=0; i<dim; i++) xx[i] = intparsXX.xx[i];
      xx[dim-1] = x; // integrate the last variable: phip for SQ, phi for PairSUB1
      integrandCUBAx(&dim, xx, &ncomp, cres, &intparsXX.pars);
      return cres[0];
    }
    double test_integrandCUBAx1_partiallyintegrated(double *xx, s_prdtv intpars, int dim) {
      s_prdtvXX intparsXX = { intpars, xx, dim };
      gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
      gsl_set_error_handler_off();
      double prec = 1e-4;
      int status;
      double res = integration_gauss(test_integrandCUBAx, 0, 1, prec, NULL, &intparsXX, &status);
      if(status) res = 0;
      gsl_set_error_handler (old_handler);
      return res;
    }


    double integral1d_DeltaFRes(s_prdtv intpars) {
      gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
      gsl_set_error_handler_off();
      double prec = 1e-4;
      int status;
      double res = integration_gauss(integrandGSLx, 0, 1, prec, NULL, &intpars, &status);
      //if(status) res = 0;
      gsl_set_error_handler (old_handler);
      return res;
    }
    double integralDd_DeltaFRes(s_prdtv intpars, int dim) {
      const int ncomp = 1;
      double cres[ncomp], error[ncomp], prob[ncomp];
      double epsrel=5e-3, epsabs=1.e-10;
      int last = 4;
      int verbose = 0;
      int nregions, neval, fail;
      //int mineval = 0, maxeval = 300000*pow(5,dim);
      int mineval = 0, maxeval = 10000*pow(5,dim);
      //int mineval = 5000, maxeval = 1000000;
      int nstart = 50000, nincrease = 20000, nbatch = 20000, gridno = 1;
      //int nstart = 20000, nincrease = 40000, nbatch = 1000, gridno = 5;
      //int nstart = 40000, nincrease = 40000, nbatch = 1000, gridno = 5;
      nregions=1; // to remove the warning
      int seed = 0;
      double result = 0;
      //int cores = 1, pmax= 10000;
      //cubacores(&cores,&pmax);
      //
      /*
      double N=1000;
      for(int j=0; j<dim; j++) {
	//if(j!=0) continue;
	cout << "--- integrand " << j << " ---" << endl;
	ostringstream ofilename;
	ofilename << "output/testintegrandPair" << j << ".dat";
	ofstream ofile(ofilename.str());
	for(int i=0; i<N; i++) {
	  double x = i/(N-1.);
	  double *xx = new double[dim];
	  for(int k=0; k<dim; k++) xx[k] = 0.2+0.1*k;
	  //xx[dim-1] = 0.00002;
	  //for(int k=0; k<dim; k++) xx[k] = 0.15*(0.05+k);
	  //xx[dim-1] = 0.0002;
	  xx[j] = x;
	  //cout << setw(4) << j << setw(8) << i << endl;
	  integrandCUBAx(&dim, xx, &ncomp, cres, &intpars);
	  //if(i%10==0) cout << cres[0] << endl;
	  //cres[0] = test_integrandCUBAx_partiallyintegrated(xx, intpars, dim);
	  ofile << setw(12) << x << setw(16) << cres[0] << endl;
	}
      }
      exit(1);
      //*/
      /*
      double N=1000;
      for(int j=0; j<dim; j++) {
	//if(j!=1) continue;
	cout << "--- integrand " << j << " ---" << endl;
	ostringstream ofilename;
	ofilename << "output/testintegrandSQ" << j << ".dat";
	ofstream ofile(ofilename.str());
	for(int i=0; i<N; i++) {
	  double x = i/(N-1.);
	  double *xx = new double[dim];
	  for(int k=0; k<dim; k++) xx[k] = 0.3*(0.15+k);
	  //xx[dim-1] = 0.00002;
	  //for(int k=0; k<dim; k++) xx[k] = 0.15*(0.05+k);
	  //xx[dim-1] = 0.0002;
	  xx[j] = x;
	  //cout << setw(4) << j << setw(8) << i << endl;
	  integrandCUBAx(&dim, xx, &ncomp, cres, &intpars);
	  //if(i%10==0) cout << cres[0] << endl;
	  //cres[0] = test_integrandCUBAx_partiallyintegrated(xx, intpars);
	  ofile << setw(12) << x << setw(16) << cres[0] << endl;
	}
      }
      exit(1);
      //*/
      //
      //Vegas(dim, ncomp, integrandCUBAx, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, nstart, nincrease, nbatch, gridno, NULL, NULL, &neval, &fail, cres, error, prob);
      //Suave(dim, ncomp, integrandCUBAx, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 1, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
      Cuhre(dim, ncomp, integrandCUBAx, &intpars, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
      result = cres[0];
      //
      if(fail == -1)  cout << "  ERROR: dimension out of range." << endl;
      else if(fail>0) cout << "  ERROR: precision not reached." << endl;
      //cout << result << " +/- " << error[0] << "   dimension: " << dim << endl;
      return result;
    }




    double newDeltaFRes(double x, double as, int nf, xTableU *tU, process proc, int dim, bool sub, double muFrat, void *pars, double rapY) {
      double b0 = beta0(nf);
      //
      s_prdtv intpars = { proc, x, as, b0, muFrat, rapY, tU, pars, nf, sub };
      double res = 0;
      if(dim==1)
	res = integral1d_DeltaFRes(intpars);
      else
	res = integralDd_DeltaFRes(intpars, dim);
      //
      return res;
    }






    // QQbar pair, triple differential
    double DeltaQQbarPairM2Yqt_Res(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      double res = 0;
      //res = newDeltaFRes(x, as, nf, tU, check1, 1, false, muFrat, pars, rapY);
      //return res;
      //res = newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_v2, 4, false, muFrat, pars, rapY);  // not particularly efficient
      //res = newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq, 4, false, muFrat, pars, rapY);  // works well, but perhaps not ideal
      // best option, stable and checked against fixed-order expansion
      res += newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_part1, 4, false, muFrat, pars, rapY);
      res += newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_part2, 4, false, muFrat, pars, rapY);
      // subtraction of fixed-order expansion here??
      //res -= newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUB1, 4, true, muFrat, pars, rapY);
      //res -= newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUB4, 4, true, muFrat, pars, rapY);
      //res -= newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUBc, 2, true, muFrat, pars, rapY);
      return res;
    }
    double DeltaQQbarPairM2Yqt_OrdAs2(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      double res = 0;
      //res = newDeltaFRes(x, as, nf, tU, check1, 2, true, muFrat, pars, rapY);
      //return res;
      //res += newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUB4, 4, true, muFrat, pars, rapY);
      //res += newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUB1, 4, true, muFrat, pars, rapY);
      //res += newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUBc, 2, true, muFrat, pars, rapY);
      res += newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_part1, 4, true, muFrat, pars, rapY);
      res += newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_part2, 4, true, muFrat, pars, rapY);
      return res;
    }
    double DeltaQQbarPairM2Yqt_ResAux(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      double res = 0;
      res += newDeltaFRes(x, as, nf, tU, QQbarPairM2YqAux, 2, false, muFrat, pars, rapY);
      //res -= newDeltaFRes(x, as, nf, tU, QQbarPairM2YqAux, 2,  true, muFrat, pars, rapY); // order alphas
      return res;
    }
    double DeltaQQbarPairM2Yqt_ResAux_OrdAs(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      double res = 0;
      res = newDeltaFRes(x, as, nf, tU, QQbarPairM2YqAux, 2,  true, muFrat, pars, rapY); // order alphas
      return res;
    }



    // QQbar pair, double differential
    double DeltaQQbarPairM2Y_Res(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      // all three implementations give the same results
      //double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2Y, 5, false, muFrat, pars, rapY);  // slow, with some errors
      //double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2Y_v2, 5, false, muFrat, pars, rapY);  // slow, cannot reach precision  // this is the integral over qt of the triple differential
      double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2Y_part1, 5, false, muFrat, pars, rapY);  // faster and precise
      return res;
    }
    double DeltaQQbarPairM2Y_Res_OrdAs2(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      //double rapY = QQbar->GetRapidity();
      // not implemented yet
      double res = 0;//newDeltaFRes(x, as, nf, tU, QQbarPairM2Y, 5, true, muFrat, pars, rapY);
      return res;
    }
    double DeltaQQbarPairM2Y_ResAux(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      if(rapY < 0) return 0;
      if(rapY > -0.5*log(x)) return 0;
      double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2YAux, 2, false, muFrat, pars, rapY);
      //double res = newDeltaFRes(x, as, nf, tU, check0, 1, false, muFrat, pars, rapY);
      return res;
    }
    double DeltaQQbarPairM2Y_ResAux_OrdAs(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      if(rapY < 0) return 0;
      if(rapY > -0.5*log(x)) return 0;
      double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2YAux, 2, true, muFrat, pars, rapY);
      return res;
    }
    double DeltaQQbarPairM2Y_ResAuxPart2(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      if(rapY < -0.5*log(2.*x)) return 0;
      if(rapY > -0.5*log(x)) return 0;
      double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2YAux_part2, 2, false, muFrat, pars, rapY);
      return res;
    }
    double DeltaQQbarPairM2Y_ResAuxPart2_OrdAs(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      if(rapY < -0.5*log(2.*x)) return 0;
      if(rapY > -0.5*log(x)) return 0;
      double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2YAux_part2, 2, true, muFrat, pars, rapY);
      return res;
    }
    double DeltaQQbarPairM2Y_LO(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2YLO, 3, false, muFrat, pars, rapY);
      return res;
    }
    double DeltaQQbarPairM2Y_LO_qqbarchannel(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarPair *QQbar = (QQbarPair*) pars;
      double rapY = QQbar->GetRapidity();
      double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2YLO_qqbarchannel, 1, false, muFrat, pars, rapY);
      return res;
    }






    // QQbar single quark, double Y-pt differential
    double DeltaQQbarSingleQuarkYqt_Res(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) pars;
      double rapY = QQbar->GetRapidity();
      //double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYq, 5, false, muFrat, pars, rapY);
      double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYq_v2, 5, false, muFrat, pars, rapY);
      return res;
    }
    double DeltaQQbarSingleQuarkYqt_ResAux(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) pars;
      double rapY = QQbar->GetRapidity();
      //double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYqAux, 2, false, muFrat, pars, rapY);  // this fails sometimes, I don't know why
      double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYqAux_v2, 2, false, muFrat, pars, rapY);  // this implementation seems to be stabler
      return res;
    }
    double DeltaQQbarSingleQuarkYqt_ResAux_OrdAs(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) pars;
      double rapY = QQbar->GetRapidity();
      //double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYqAux, 2, true, muFrat, pars, rapY);
      double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYqAux_v2, 2, true, muFrat, pars, rapY);
      return res;
    }

    // QQbar single quark, Y differential
    double DeltaQQbarSingleQuarkY_Res(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) pars;
      double rapY = QQbar->GetRapidity();
      double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkY, 1, false, muFrat, pars, rapY);  /// fix dimension!
      return res;
    }
    double DeltaQQbarSingleQuarkY_ResAux(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) pars;
      double rapY = QQbar->GetRapidity();
      double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYAux, 1, false, muFrat, pars, rapY);  /// fix dimension!
      return res;
    }


  };
};

