#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <cstdlib>
#include <cuba.h>
#include <gsl/gsl_errno.h>

#include "CoefficientFunctionsMultiDiff.hh"
#include "QQbarPair.hh"
#include "Pqg.hh"
#include "math/borel.hh"
#include "math/special_functions.hh"
#include "integration.hh"
#include "version.hh"
#include "expansion_AD_SF.hh"
//#include <complex_bessel/sp_airy.h>


using namespace std;



namespace HELL {

  extern double phinonprime(double xi, double xi2, double phiprime);


  namespace HELLdiff {


  //
  // integrand of the form:  U'(x, xi, muF) * Coff(xi, pars)   or   - U(x, xi, muF) * dCoff/dxi(xi, pars)
  //
  double integrand(double xi, double x, xTableU *tU, process p, double asb0, double muFrat, void *par) { //, double *intVars) {
    double dU = 0;
    double xiF = muFrat*muFrat;  // needed for the collinear subtractions
    double xiU = xi/xiF;  // use the variable ( Q^2/muF^2 * xi )
    double  U = tU->evalU(x, xiU, dU);
    /*
    if(xiU<1) {  // apply higher-twist damping
      int kk = 1;
      double  HTdamp = 1 - pow(-asb0*log(xiU),kk*(1+1/asb0));
      double dHTdamp = (HTdamp-1) * kk*(1+1/asb0)/xiU/log(xiU);
      dU = dU * HTdamp + U * dHTdamp;
      U *= HTdamp;
    }
    dU /= xiF;    // convert derivatives of U wrt their arguments (xiU) to derivatives wrt xi
    //*/
    //
    //cout << x << "  " << xi << endl;
    double res = U;
    //
    if(p == QQbarPairM2YqAux) {
      //QQbarPair *QQbar = (QQbarPair*) par;
      res *= 1;//QQbar->diffXS_M_Y_pt_Aux(xi);
    }
    //
    //cout << setw(15) << xi << setw(15) << x1 << setw(35) << res << endl;
    return res;
  }


  //
  // integrand of the form:  U'(x1, xi1, muF) * U'(x2, xi2, muF) * Coff(xi1, xi2, pars)
  //
  //double integrand2d(double t, double y, double x1, double x2, xTableU *tU, process p, double asb0, double muFrat, void *par, double *intVars) {
  double integrand2d(double xi1, double xi2, double x1, double x2, xTableU *tU, process p, double asb0, double muFrat, void *par, double *intVars, double phip=0, double xi=0, double eta=0) {
    //double xi1 = t*(1+y);
    //double xi2 = t*(1-y);
    double xiF = muFrat*muFrat;
    double xiU1 = xi1/xiF;
    double xiU2 = xi2/xiF;
    /////////////////////////////////////////////////////
    double xi0 = exp(-1./asb0);
    if(xiU1<xi0) {
      if(fabs(xiU1-xi0)/xi0<1e-3) {
	xiU1 = xi0;
      } else {
	//cout << "cazzo!  xi0=" << xi0 << "  xi1-xi0=" << xiU1-xi0 << endl;//<< "   t="<<t << " y="<<y << endl;
	return 0;
      }
    }
    if(xiU2<xi0) {
      if(fabs(xiU2-xi0)/xi0<1e-3) {
	xiU2 = xi0;
      } else {
	//cout << "cazzo!  xi0=" << xi0 << "  xi2-xi0=" << xiU2-xi0 << endl;// << "   xi2="<<xiU2 << "   t="<<t << " y="<<y << endl;
	return 0;
      }
    }
    if(xiU1<xi0 || xiU2<xi0) return 0;
    /////////////////////////////////////////////////////
    //
    double res = 0;
    //
    if(p == QQbarPairM2Yq || p == QQbarPairM2Y_v2) {
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xrescale = sqrt(1+qthat*qthat);
      if(xrescale!=xrescale) cout << "What?!" << endl;
      x1 *= xrescale;
      x2 *= xrescale;
      if(x1>1 || x2>1) return 0;
      double dU1;  tU->evalU(x1, xiU1, dU1);
      double dU2;  tU->evalU(x2, xiU2, dU2);
      double dU1a; tU->evalU(x2, xiU1, dU1a); // swapped x1 and x2
      double dU2a; tU->evalU(x1, xiU2, dU2a);
      //res = dU1 * dU2 * QQbar->diffXS_M_Y_pt(xi1, xi2, intVars[2], intVars[3]) + dU1a * dU2a * QQbar->diffXS_M_Y_pt(xi2, xi1, intVars[2], intVars[3]);
      res = (dU1*dU2 + dU1a*dU2a) * QQbar->diffXS_M_Y_pt(xi1, xi2, intVars[2], intVars[3]);
      if(res!=res) {
	cout << res << "  " << x1 << "  " << x2 << "  " << xi1 << "  " << xi2 << "  " << intVars[0] << "  " << intVars[1] << "  " << intVars[2] << "  " << intVars[3] << "  " << dU1 << "  " << dU2 << "  " << QQbar->diffXS_M_Y_pt(xi1, xi2, intVars[2], intVars[3]) << endl;
	res = 0;
      }
    }
    if(p == QQbarPairM2Yq_v2) {
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xrescale = sqrt(1+qthat*qthat);
      if(xrescale!=xrescale) cout << "What?!" << endl;
      x1 *= xrescale;
      x2 *= xrescale;
      if(x1>1 || x2>1) return 0;
      double dU1; tU->evalU(x1, xiU1, dU1);
      double dU2; tU->evalU(x2, xiU2, dU2);
      //double cosphipmax = (xi2+qthat*qthat-xi0)/2./qthat/sqrt(xi2);
      //double phipmin = (cosphipmax>1? 0 : acos(cosphipmax));
      //double phip = phipmin + 2*(M_PI-phipmin)*intVars[1];
      res = dU1*dU2 * QQbar->diffXS_M_Y_pt_v2(xi2, phip, intVars[2], intVars[3]);
    }
    if(p == QQbarPairM2Y) {
      QQbarPair *QQbar = (QQbarPair*) par;
      double jac, phi = QQbar->intphi(intVars[4], xi1, xi2, jac);
      double xrescale = sqrt(1+xi1+xi2+2*sqrt(xi1*xi2)*cos(phi));
      if(xrescale!=xrescale) cout << xi1 << "  " << xi2 << "  " << intVars[4] << "  " << phi << endl;
      x1 *= xrescale;
      x2 *= xrescale;
      if(x1>1 || x2>1) return 0;
      //cout << xrescale << endl;
      double dU1;  tU->evalU(x1, xiU1, dU1);
      double dU2;  tU->evalU(x2, xiU2, dU2);
      //double dU1a; tU->evalU(x2, xiU1, dU1a); // swapped x1 and x2
      //double dU2a; tU->evalU(x1, xiU2, dU2a);
      //res = dU1 * dU2 * QQbar->diffXS_M_Y(xi1, xi2, intVars[2], intVars[3], intVars[4]) + dU1a * dU2a * QQbar->diffXS_M_Y(xi2, xi1, intVars[2], intVars[3], intVars[4]);
      //res = (dU1*dU2 + dU1a*dU2a) * QQbar->diffXS_M_Y(xi1, xi2, intVars[2], intVars[3], intVars[4]);
      res = dU1*dU2 * QQbar->diffXS_M_Y(xi1, xi2, intVars[2], intVars[3], intVars[4]);
      //cout << res << endl;
      if(res!=res) {
	cout << res << "  " << x1 << "  " << x2 << "  " << xi1 << "  " << xi2 << "  " << intVars[0] << "  " << intVars[1] << "  " << intVars[2] << "  " << intVars[3] << "  " << dU1 << "  " << dU2 << "  " << QQbar->diffXS_M_Y(xi1, xi2, intVars[2], intVars[3], intVars[4]) << endl;
	res = 0;
      }
    }
    if(p == QQbarPairM2YAux) {
      QQbarPair *QQbar = (QQbarPair*) par;
      if(x1>1) return 0;
      //cout << xrescale << endl;
      double dU1;  tU->evalU(x1, xiU1, dU1);
      res = dU1 * QQbar->diffXS_M_Y_Aux(intVars[0], intVars[1]);
      //cout << res << endl;
      if(res!=res) {
	cout << res << "  " << x1 << "  " << xi1 << "  " << intVars[0] << "  " << intVars[1] << "  " << dU1 << "  " << QQbar->diffXS_M_Y_Aux(intVars[0], intVars[1]) << endl;
	res = 0;
      }
    }
    if(p == QQbarPairM2YLO) {
      QQbarPair *QQbar = (QQbarPair*) par;
      res = QQbar->diffXS_M_Y_LO(intVars[0], intVars[1], intVars[2]);
      if(res!=res) {
	//cout << res << "  " << intVars[0] << "  " << intVars[1] << "  " << QQbar->diffXS_M_Y_LO(intVars[0], intVars[1]) << endl;
	res = 0;
      }
    }
    //
    //
    // SINGLE QUARK
    //
    if(p == QQbarSingleQuarkYq) {
      if(x1>1 || x2>1) return 0;
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double z = QQbar->GetX()/x1/x2;
      double y = QQbar->GetRapidity() - 0.5*log(x1/x2);
      double dU1;  tU->evalU(x1, xiU1, dU1);
      double dU2;  tU->evalU(x2, xiU2, dU2);
      double dU1a; tU->evalU(x2, xiU1, dU1a); // swapped x1 and x2
      double dU2a; tU->evalU(x1, xiU2, dU2a);
      res = dU1 * dU2 * QQbar->diffXS_Y_pt(xi1, xi2, z, y, intVars[4]) + dU1a * dU2a * QQbar->diffXS_Y_pt(xi2, xi1, z, y, intVars[4]);
      //res = (dU1*dU2 + dU1a*dU2a) * QQbar->diffXS_Y_pt(xi1, xi2, z, y, intVars[4]);
      res /= z;
    }
    if(p == QQbarSingleQuarkYq_v2) {
      if(x1>1 || x2>1) return 0;
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double z = QQbar->GetX()/x1/x2;
      //cout << "   z = " << z<< endl;
      //double etaa = QQbar->GetRapidity() - 0.5*log(x1/x2);
      //cout << etaa-eta << endl;
      double dU1; tU->evalU(x1, xiU1, dU1);
      double dU2; tU->evalU(x2, xiU2, dU2);
      //double cosphipmax = (xi2+xi-xi0)/2./sqrt(xi*xi2);
      //double phipmin = (cosphipmax>1? 0 : acos(cosphipmax));
      //double phip = phipmin + 2*(M_PI-phipmin)*intVars[1];
      //double xi1test = xi+xi2-2*sqrt(xi*xi2)*cos(phip);
      //cout << xi1test << "  " << xi1 << endl;
      //cout << "dopo: " << phip << endl;
      res = dU1 * dU2 * QQbar->diffXS_Y_pt_v2(xi, xi2, z, eta, phip, xi1);
      res /= z;
      if(res != res) {
	cout << res << "  " << xi1 << "  " << phip << "  " << xi << "  " << xi2 << "  " << z << "  " << eta << "  " << dU1 << "  " << dU2 << "  " << phinonprime(xi,xi2,phip) << endl;
	res = 0;
      }
    }
    //
    //cout << res << endl;
    return res;
  }


  double integrand2d_SUB(double xi1, double xi2, double x1, double x2, process p, double asb0, double muFrat, void *par, double *intVars, int nf, double as) {
    double res = 0;
    //
    if(p == QQbarPairM2Yq_SUB4) {
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xrescale = sqrt(1+qthat*qthat);
      if(xrescale!=xrescale) cout << "What?!" << endl;
      x1 *= xrescale;
      x2 *= xrescale;
      if(x1>1 || x2>1) return 0;
      double xP0x1 = x1*P0(x1,nf);
      double xP0x2 = x2*P0(x2,nf);
      res = as*as* xP0x1 * xP0x2 /xi1 /xi2 * QQbar->diffXS_M_Y_pt(xi1, xi2, intVars[2], intVars[3]);
      res *= pow(1-x1,2); // damping!!
      res *= pow(1-x2,2); // damping!!
    }
    if(p == QQbarPairM2Yq_SUB1) {
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xrescale = sqrt(1+qthat*qthat);
      if(xrescale!=xrescale) cout << "What?!" << endl;
      x1 *= xrescale;
      x2 *= xrescale;
      if(x1>1 || x2>1) return 0;
      double xP0x1 = x1*P0(x1,nf);
      double xP0x2 = x2*P0(x2,nf);
      res = as*as* xP0x1 * xP0x2 /xi1 * QQbar->diffXS_M_Y_pt_SUB(xi1, intVars[1], intVars[2], intVars[3]);
      res *= pow(1-x1,2); // damping!!
      res *= pow(1-x2,2); // damping!!
    }
    if(p == QQbarPairM2Yq_SUBc) {
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double qthat2 = qthat*qthat;
      if(qthat2>=1) return 0;
      double xrescale = sqrt(1+qthat*qthat);
      if(xrescale!=xrescale) cout << "What?!" << endl;
      x1 *= xrescale;
      x2 *= xrescale;
      if(x1>1 || x2>1) return 0;
      double xP0x1 = x1*P0(x1,nf);
      double xP0x2 = x2*P0(x2,nf);
      res = as*as* xP0x1 * xP0x2 * QQbar->diffXS_M_Y_pt_SUBc(intVars[0], intVars[1]);
      res *= pow(1-x1,2); // damping!!
      res *= pow(1-x2,2); // damping!!
    }
    //
    return res;
  }







  struct s_prdtv { process p; double d1, d2, d3, d4, d5; xTableU *tU; void *par; int nf; bool sub; };
  double integrandGSLx(double x, void *pars) {
    s_prdtv p = *(s_prdtv*)pars;
    double xx  = p.d1;
    //xTableU *tU = p.tU;
    double a0 = p.d2;  // as(Q^2)
    double b0 = p.d3;  // beta0
    double muFrat = p.d4;
    //double lFQ = 2*log(muFrat);
    double xiF = muFrat*muFrat;
    //double xi0 = real(r); // may be negative - problematic
    //double xi0 = a0*b0;
    //if(real(r)>0 && real(r)>xi0) xi0 = real(r); // not sure if it makes sense...
    //double xi = exp(-1./a0/b0) + (1.-x)/x;  double jac = 1./x/x;
    //double mult = 3; double t = (mult*x-1)/a0/b0; double xi = exp(t);  double jac = xi*mult/a0/b0;
    //double mult = 3; double u0=1./a0/b0-lFQ; double t = x*u0; double xi1 = exp(-t); double xi2 = exp(mult*t); double jac1 = xi1*u0; double jac2 = mult*xi2*u0;
    //double mult = 3; if(a0<0.007) mult=2; double u01=1./a0/b0-lFQ; double t1 = x*u01; double u02=1./a0/b0; double t2 = x*u02; double xi1 = exp(-t1); double xi2 = exp(mult*t2); double jac1 = xi1*u01; double jac2 = mult*xi2*u02;
    double mult = 3; if(a0<0.007) mult=2; double u0=1./a0/b0; double t = x*u0; double xi1 = xiF*exp(-t); double xi2 = xiF*exp(mult*t); double jac1 = xi1*u0; double jac2 = mult*xi2*u0;
    //double mult = 3; double u=x*x; double t = u/a0/b0; double xi1 = exp(-t); double xi2 = exp(mult*t); double jac1 = xi1/a0/b0*2*x; double jac2 = mult*xi2/a0/b0*2*x;
    //double amin = 0.01*a0, amax = 2000*a0;  double a = amin + (amax-amin)*x;  double xi = exp((1/a-1/a0)/b0);  double jac = -xi/(-b0*a*a) * (amax-amin);
    //double Xmin = 0.1, Xmax = 80;  double X = Xmin + (Xmax-Xmin)*x;  double xi = exp((1-X)/X/xi0);  double jac = xi/X/X/xi0*(Xmax-Xmin);
    //double Xmin = 0.05, Xmax = 70;  double X1 = Xmin + (1-Xmin)*x;  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin); double X2 = 1 + (Xmax-1)*(1-x);  double xi2 = exp((1-X2)/X2/xi0);  double jac2 = xi2/X2/X2/xi0*(Xmax-1);
    //double Xmin = 0.1;  double X1 = Xmin + (1-Xmin)*x;  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin);  double X2 = x;  double xi2 = exp((1-1/X2)*X2/xi0);  double jac2 = xi2/xi0;
    //cout << x << "  " << xi << endl;
    //cout << xi1 << "   " << xi2 << endl;
    //cout << u01 << "  " << lFQ << endl;
    //if(1+real(r)*log(xi1)<0) cout << "warning! xi1="<< xi1 << "  r="<< r << endl;
    //if(1+real(r)*log(xi2)<0) cout << "warning! xi2="<< xi2 << "  r="<< r << endl;
    //
    //cout << xi0 << "  " << a0 << "  "  << b0 << "  " << xi2 << "  " << x << endl;
    //if(x>1e-5) cout << xi0 << "  " << a0 << "  "  << b0 << "  " << xi << "  " << x << endl;
    //double result = jac  * integrand(xi,  xx, p.tU, p.p, a0*b0, muFrat, p.par);
    double result = jac1 * integrand(xi1, xx, p.tU, p.p, a0*b0, muFrat, p.par) + jac2 * integrand(xi2, xx, p.tU, p.p, a0*b0, muFrat, p.par);
    //if(result==result) cout << result << endl;
    //
    //cout << setw(15) << xi << setw(15) << x[1] << setw(35) << res << endl;
    //delete tU;
    return result;
  }
  int integrandCUBAx(const int *ndim, const double x[], const int *ncomp, double res[], void *pars, int whichFunction) {
    //if(x[1]==0)
    s_prdtv p = *(s_prdtv*)pars;
    double xx  = p.d1;
    double xz  = xx;
    double a0 = p.d2;  // as(Q^2)
    double b0 = p.d3;  // beta0
    double muFrat = p.d4;
    double rapY  = p.d5;
    double phip = 0;
    double xi = 0;
    double eta = 0;
    //double lFQ = 2*log(muFrat);
    double xiF = muFrat*muFrat;
    void *par = p.par;
    process pr = p.p;
    double xi1a=1, xi2a=1, xi1b=1, xi2b=1, jac1=1, jac2=1;
    //
    double *intVars = new double[*ndim];
    for(int i=0; i<*ndim; i++) intVars[i] = x[i];
    //
    if(pr == QQbarPairM2Yq) {
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double logxi0 = log(xiF)-1./a0/b0;
      double xi0 = exp(logxi0);
      //double xi0 = xiF*exp(-1./a0/b0);
      double mult=1.2;
      //*
      //double tmin=max(qthat*qthat/4,xi0);  double tmax=min(max(xiF*pow(xiF/xi0,mult),100*qthat*qthat),1e5);
      //double u=x[0]*log(tmax/tmin);    double t1=tmin*exp(u);     jac1*=t1*log(tmax/tmin);
      //double t1=exp((1-x[0])*log(tmin)+x[0]*log(tmax));     jac1*=t1*log(tmax/tmin);
      //double y1max=min(qthat/2,sqrt(t1)-sqrt(xi0)); double y1=y1max*x[1]; jac1*=y1max;
      double tmin=qthat*qthat/4;  double tmax=min(max(xiF*pow(xiF/xi0,mult),100*qthat*qthat),1e5);
      //double u=x[0]*log(tmax/tmin);    double t1=tmin*exp(u);     jac1*=t1*log(tmax/tmin);
      double t1=exp((1-x[0])*log(tmin)+x[0]*log(tmax));     jac1*=t1*log(tmax/tmin);
      double y1max=qthat/2; double y1=y1max*(1-x[1]*x[1]*x[1]); jac1*=3*x[1]*x[1]*y1max;
      /*/
      double y1max=qthat/2; double y1=y1max*(1-x[1]); jac1*=y1max;
      double tmin=max(qthat*qthat/4,pow(exp(logxi0/2)+y1,2));  double tmax=min(max(xiF*pow(xiF,mult)*exp(-mult*logxi0),100*qthat*qthat),1e5);
      double u=x[0]*log(tmax/tmin);  //jac1 *= 2*x[0];
      double t1=tmin*exp(u);     jac1*=t1*log(tmax/tmin);
      //*/
      //double t2=xiF*exp(mult*u); jac2*=mult*t2*u0; double y2max=min(qthat/2,sqrt(t2)-sqrt(xi0)); double y2=y2max*x[1]; jac2*=y2max;
      //xi1a = t1*pow(1+y1/sqrt(t1),2);  xi2a = t1*pow(1-y1/sqrt(t1),2);  jac1*=4*sqrt(t1)*(1-y1*y1/t1);
      xi1a = pow(sqrt(t1)+y1,2);  xi2a = pow(sqrt(t1)-y1,2);  jac1*=4*sqrt(t1)*(1-y1*y1/t1);
      //xi1a = t1+y1*y1+2*y1*sqrt(t1);  xi2a = t1+y1*y1-2*y1*sqrt(t1);  jac1*=4*sqrt(t1)*(1-y1*y1/t1);
      //if(xi2a<xi0) cout << y1max << "   " << y1 << "   " << xi1a << "   " << xi2a << "   " << xi0 << endl;
      //xi1b = t2*pow(1+y2/sqrt(t2),2);  xi2b = t2*pow(1-y2/sqrt(t2),2);  jac2*=4*sqrt(t2)*(1-y2*y2/t2);
      //double sqrtarg = 4*xi1a*xi2a-(qthat*qthat-xi1a-xi2a)*(qthat*qthat-xi1a-xi2a);
      //if(sqrtarg<0) {
      //  cout << ">>>>>Warning: argument of sqrt is negative: " << sqrtarg << endl;
      //}
    } else if(pr == QQbarPairM2Yq_SUB4) {
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xi0=1e-6; //xiF*exp(-1./a0/b0);
      double mult=1.2;
      double tmin=max(qthat*qthat/4,1.);  double tmax=max(xiF*pow(xiF/xi0,mult),100*qthat*qthat);
      double u=x[0]*log(tmax/tmin);
      double t1=tmin*exp(u);     jac1*=t1*log(tmax/tmin);      double y1max=min(qthat/2,sqrt(t1)-1.); double y1=y1max*x[1]; jac1*=y1max;
      xi1a = t1*pow(1+y1/sqrt(t1),2);  xi2a = t1*pow(1-y1/sqrt(t1),2);  jac1*=4*sqrt(t1)*(1-y1*y1/t1);
    } else if(pr == QQbarPairM2Yq_SUB1) {
      double xi0=1e-6;//xiF*exp(-1./a0/b0);
      //xi0 *= 5;
      //double u = x[0];
      double u = x[0]*x[0]; jac1 *= 2*x[0];
      //double u = x[0]*x[0]*x[0]; jac1 *= 3*x[0]*x[0];
      //double u = exp(-1/x[0]); jac1 *= u;
      xi1a = xi0+(1-xi0)*u;  xi2a = 0;  jac1*=1-xi0;
    } else if(pr == QQbarPairM2Yq_SUBc) {
      xi1a = 0;  xi2a = 0;
    } else if(pr == QQbarPairM2Yq_v2) {
      QQbarPair *QQbar = (QQbarPair*) par;
      double qthat = QQbar->GetQthat();
      double xi0=xiF*exp(-1./a0/b0); double mult=2;
      double tmin=xi0;  double tmax=xiF*pow(xiF/xi0,mult);
      double u=x[0]*log(tmax/tmin);
      double t1=tmin*exp(u);     jac1*=t1*log(tmax/tmin);
      double cosphipmax = (t1+qthat*qthat-xi0)/2./qthat/sqrt(t1);
      double phipmin = (cosphipmax>1? 0 : myacos(cosphipmax));
      phip=phipmin + 2*(M_PI-phipmin)*x[1];   jac1*=2*(M_PI-phipmin);
      xi1a = t1+qthat*qthat-2*qthat*sqrt(t1)*cos(phip);  xi2a = t1;
    }
    //
    else if(pr == QQbarPairM2Y) {
      /*
      double logxi0 = log(xiF)-1./a0/b0;
      double xi0 = exp(logxi0);
      double mult=1.2;
      double tmin=xi0;  double tmax=min(xiF*pow(xiF/xi0,mult),1e5);
      double t1=exp((1-x[0])*log(tmin)+x[0]*log(tmax));     jac1*=t1*log(tmax/tmin);
      double y1max=sqrt(t1)-exp(logxi0/2); double y1=y1max*(1-x[1]*x[1]); jac1*=2*x[1]*y1max;
      xi1a = pow(sqrt(t1)+y1,2);  xi2a = pow(sqrt(t1)-y1,2);  jac1*=4*sqrt(t1)*(1-y1*y1/t1);
      /*/
      double logxi0 = log(xiF)-1./a0/b0;
      double xi0 = exp(logxi0);
      double mult=1.2;
      double tmin=xi0;  double tmax=min(xiF*pow(xiF/xi0,mult),1e5);
      xi1a = exp((1-x[0])*log(tmin)+x[0]*log(tmax));  xi2a = exp((1-x[1])*log(tmin)+x[1]*log(tmax));  jac1*=xi1a*xi2a*log(tmax/tmin)*log(tmax/tmin);
      //*/
    } else if(pr == QQbarPairM2Y_v2) {
      QQbarPair *QQbar = (QQbarPair*) par;
      //double qthat = x[4]/(1-x[4]);
      double qtmax = sqrt(exp(-2*fabs(QQbar->GetRapidity()))/xx - 1);
      double qthat = x[4]*qtmax;
      QQbar->SetKinematics(QQbar->GetX(), QQbar->GetInvariantMass(), QQbar->GetRapidity(), qthat*QQbar->GetInvariantMass());
      double xi0=xiF*exp(-1./a0/b0); double mult=1.2;
      double tmin=max(qthat*qthat/4,xi0);  double tmax=max(xiF*pow(xiF/xi0,mult),100*qthat*qthat);
      double u=x[0]*log(tmax/tmin);
      double t1=tmin*exp(u);     jac1*=t1*log(tmax/tmin);      double y1max=min(qthat/2,sqrt(t1)-sqrt(xi0)); double y1=y1max*x[1]; jac1*=y1max;
      //double t2=xiF*exp(mult*u); jac2*=mult*t2*u0; double y2max=min(qthat/2,sqrt(t2)-sqrt(xi0)); double y2=y2max*x[1]; jac2*=y2max;
      xi1a = t1*pow(1+y1/sqrt(t1),2);  xi2a = t1*pow(1-y1/sqrt(t1),2);  jac1 *= 4*sqrt(t1)*(1-y1*y1/t1);
      //xi1b = t2*pow(1+y2/sqrt(t2),2);  xi2b = t2*pow(1-y2/sqrt(t2),2);  jac2*=4*sqrt(t2)*(1-y2*y2/t2);
      //jac1 *= 2*qthat/(1-x[4])/(1-x[4]);
      jac1 *= 2*qthat*qtmax;
    }
    //
    // SINGLE QUARK
    //
    else if(pr == QQbarSingleQuarkYq) {
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double pthat = QQbar->GetPthat();
      double intz = xx+(1-xx)*x[3];
      rapY  = -log(xx/intz) * (x[2]-1./2.); // this is yhat, the integration variable
      jac1 *= -log(xx/intz) * (1-xx);
      jac2 *= -log(xx/intz) * (1-xx);
      xz = xx/intz;
      eta = QQbar->GetRapidity() - rapY;
      double sqrtstuff = sqrt(pthat*pthat + 1/intz - sqrt(1+pthat*pthat)*(exp(eta)+exp(-eta))/sqrt(intz));
      double qthatmax = sqrtstuff + pthat;
      double qthatmin = max( sqrtstuff - pthat, 0.);
      double xi0=xiF*exp(-1./a0/b0); double mult=1.2;
      double tmin=max(qthatmin*qthatmin/4,xi0);  double tmax=max(xiF*pow(xiF/xi0,mult),100*qthatmin*qthatmin);
      double u=x[0]*log(tmax/tmin);
      double t1=tmin*exp(u);     jac1*=t1*log(tmax/tmin);
      double y1max=min(qthatmax/2,sqrt(t1)-sqrt(xi0)); double y1=y1max*x[1]; jac1*=y1max;
      //double t2=xiF*exp(mult*u); jac2*=mult*t2*u0; double y2max=min(qthat/2,sqrt(t2)-sqrt(xi0)); double y2=y2max*x[1]; jac2*=y2max;
      xi1a = t1*pow(1+y1/sqrt(t1),2);  xi2a = t1*pow(1-y1/sqrt(t1),2);  jac1*=4*sqrt(t1)*(1-y1*y1/t1);
      //xi1b = t2*pow(1+y2/sqrt(t2),2);  xi2b = t2*pow(1-y2/sqrt(t2),2);  jac2*=4*sqrt(t2)*(1-y2*y2/t2);
    } else if(pr == QQbarSingleQuarkYq_v2) {
      QQbarSingleQuark *QQbar = (QQbarSingleQuark*) par;
      double pthat = QQbar->GetPthat();
      double y = QQbar->GetRapidity(); // this is y, the argument of the on-shell coefficient
      //cout << y << endl;
      //double exp2yhatmax = min( 1./xx, (1-exp(y)*sqrt(xx*(1+pthat*pthat)))/(-xx*pthat*pthat+exp(-y)*sqrt(xx*(1+pthat*pthat))) );
      //double exp2yhatmin = max( xx, (-xx*pthat*pthat+exp(y)*sqrt(xx*(1+pthat*pthat)))/(1-exp(-y)*sqrt(xx*(1+pthat*pthat))) );
      if((-xx*pthat*pthat+exp(-y)*sqrt(xx*(1+pthat*pthat)))<0) cout << "AAAAAAAAA" << endl;
      if((-xx*pthat*pthat+exp( y)*sqrt(xx*(1+pthat*pthat)))<0) cout << "BBBBBBBBB" << endl;
      double A = sqrt(xx*(1+pthat*pthat));
      //double exp2yhatmax = (1-exp(y)*sqrt(xx*(1+pthat*pthat)))/(-xx*pthat*pthat+exp(-y)*sqrt(xx*(1+pthat*pthat)));
      //double exp2yhatmin = (-xx*pthat*pthat+exp(y)*sqrt(xx*(1+pthat*pthat)))/(1-exp(-y)*sqrt(xx*(1+pthat*pthat)));
      double exp2yhatmax = (1-A*exp(y))/A*exp(y);
      double exp2yhatmin = A*exp(y)/(1-A*exp(-y));
      //cout << exp2yhatmax << "  " << exp2yhatmin << endl;
      if(exp2yhatmax < exp2yhatmin) cout << "Problem: yhat limits inverted" << endl;
      rapY  = 0.5*log(exp2yhatmin) + 0.5*(log(exp2yhatmax)-log(exp2yhatmin))*x[2]; // this is yhat, the integration variable
      //rapY = -1.69;
      //cout << rapY << endl;
      jac1 *= 0.5*(log(exp2yhatmax)-log(exp2yhatmin));
      eta = y - rapY;  // this is eta = y-yhat, the argument of the off-shell coefficient
      double B = sqrt(1+pthat*pthat) * cosh(eta);
      double zmin = xx*exp(2*fabs(rapY));
      //double zmax = 1./pow(B+sqrt(B*B-pthat*pthat),2);
      double zmax = 1./4./B/B; // new limit found on board
      if(zmax < zmin) {
	if(zmin-zmax>1e-10) cout << "Problem: z limits inverted: " << zmin << " < " << zmax << "   difference: " << zmin-zmax << endl;
	else zmin = zmax;
      }
      double intz = zmin+(zmax-zmin)*(1+cos(M_PI*x[3]))/2;
      //cout << "intz = " << intz << endl;
      jac1 *= (zmax-zmin)*M_PI*sin(M_PI*x[3])/2;
      xz = xx/intz;
      double stuff = pthat*pthat + 1/intz - 2*B/sqrt(intz);
      double sqrtstuff = sqrt(pthat*pthat + 1/intz - 2*B/sqrt(intz));
      if(stuff<0 && stuff>-1e-10) sqrtstuff = 0;
      if(sqrtstuff != sqrtstuff) cout << "NaN again... " << pthat*pthat + 1/intz - 2*B/sqrt(intz) << endl;
      //cout << pthat*pthat + 1/intz - sqrt(1+pthat*pthat)*(exp(eta)+exp(-eta))/sqrt(intz) << endl;
      double qthatmax = sqrtstuff + pthat; //min( sqrtstuff + pthat, sqrt(1./intz-4));
      double qthatmin = max( sqrtstuff - pthat, 0.);
      if(qthatmax < qthatmin) cout << "Problem: xi limits inverted" << endl;
      double xi0=xiF*exp(-1./a0/b0); double mult=2;//0.2;
      double t1min=qthatmin*qthatmin;  double t1max=qthatmax*qthatmax;
      double x0 = x[0], epsilon = 1e-5;
      if(x[0]<  epsilon) x0 =   epsilon;
      if(x[0]>1-epsilon) x0 = 1-epsilon;
      //double u1=x[0]*log(t1max/t1min);
      //double t1=t1min*exp(u1);     jac1*=t1*log(t1max/t1min);  // xi
      //double t1=t1min + (t1max-t1min)*x0;     jac1*=(t1max-t1min);  // xi
      double t1=t1min + (t1max-t1min)*(1+cos(M_PI*x0))/2;     jac1*=(t1max-t1min)*M_PI*sin(M_PI*x0)/2;  // xi
      //if(t1 != t1) cout << "NaN ...  " << t1max << "  " << t1min << "  " << t1 << endl;
      double t2min=xi0;  double t2max=xiF*pow(xiF/xi0,mult);
      //double t2=t2min + (t2max-t2min)*x[1];     jac1*=(t2max-t2min); // xi2
      //double u2=x[1]*log(t2max/t2min);     double t2=t2min*exp(u2);     jac1*=t2*log(t2max/t2min); // xi2
      double yxi=max(1e-10,log(t1/t2min)/log(t2max/t2min));  double yxi2=pow(yxi*pow(1-yxi,2),1./3.);  double axi=yxi-yxi*(1-yxi)/yxi2+yxi2;  double y2 = (pow(x[1]*x[1]-axi,3)+pow(axi,3))/(pow(1-axi,3)+pow(axi,3));
      double u2=y2*log(t2max/t2min);     double t2=t2min*exp(u2);     jac1*=t2*log(t2max/t2min)*3*pow(x[1]*x[1]-axi,2)/(pow(1-axi,3)+pow(axi,3))*2*x[1]; // xi2
      double cosphipmax = (t2+t1-xi0)/2./sqrt(t1*t2);
      //if(cosphipmax<1) cout << "Cosphipmax: " << cosphipmax << "  " << t2 << endl;
      double phipmin = (cosphipmax>1? 0 : myacos(cosphipmax));
      phip = phipmin + 2*(M_PI-phipmin)*x[4];   jac1*=2*(M_PI-phipmin);
      //phip = phipmin + 2*(M_PI-phipmin)*0.99;   jac1*=2*(M_PI-phipmin);///////////////
      xi = t1;
      //cout << "prim: " << phip << endl;
      xi1a = t1+t2-2*sqrt(t1*t2)*cos(phip);  xi2a = t2;
      //cout << x[1] << "  " << t2 << endl;
      if(t2!=t2) cout << yxi << "  " << axi << "  " << xi << "  " << t2 << "  " << x[1] << "  " << phipmin << "  " << xi1a << "  " << xi0 << endl;
    }
    //
    //
    /*
    else {
      double u0=1./a0/b0; double xi0=xiF*exp(-u0); double u=x[0]*u0; double mult=1.2;
      //double t1=xiF*exp(-u);     jac1*=t1*u0;      double y1=(1-xi0/t1)*(1.-x[1]*x[1]); jac1*=(1-xi0/t1)*2*x[1];
      double t1=xiF*exp(-u);     jac1*=t1*u0;      double y1=(1-xi0/t1)*x[1]; jac1*=(1-xi0/t1);
      //double t2=xiF*exp(mult*u); jac2*=mult*t2*u0; double y2=(1-xi0/t2)*(1.-x[1]*x[1]); jac2*=(1-xi0/t2)*2*x[1]; // is it better to unlink the two integration variables?
      double t2=xiF*exp(mult*u); jac2*=mult*t2*u0; double y2=(1-xi0/t2)*x[1]; jac2*=(1-xi0/t2); // is it better to unlink the two integration variables?
      xi1a = t1*(1+y1);  xi2a = t1*(1-y1);  jac1*=2*t1;
      xi1b = t2*(1+y2);  xi2b = t2*(1-y2);  jac2*=2*t2;
    }
    */
    //
    if(pr == ggH) { // inclusive
      rapY  = -log(xx) * (x[2]-1./2.);
      jac1 *= -log(xx);
      jac2 *= -log(xx);
    }
    //
    //cout << rapY << endl;
    // these are the x arguments of the evolutors U
    double xx1  = sqrt(xz)*exp( rapY);
    double xx2  = sqrt(xz)*exp(-rapY);
    //cout << xx1 << "  " << xx2 << endl;
    //
    // if(pr == ggH) { // inclusive ---- ALTERNATIVE IMPLEMENTATION
    //   xx1  = xx+(1-xx)*x[2];
    //   xx2  = xx/xx1;
    //   jac1 *= (1-xx)/xx1;
    //   jac2 *= (1-xx)/xx1;
    // }
    if(pr == QQbarPairM2YAux) {
      //QQbarPair *QQbar = (QQbarPair*) par;
      xx1 = exp(2*rapY);
      xi1a = xx1/xx-1;
    }
    //
    if(whichFunction == 0) {
      res[0] = jac1 * integrand2d(xi1a, xi2a, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars, phip, xi, eta) + jac2 * integrand2d(xi1b, xi2b, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars, phip, xi, eta);
      //res[0] = jac1 * integrand2d(t1, y1, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars) + jac2 * integrand2d(t2, y2, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars);
    } else if(whichFunction == 1) {
      if(!p.sub) res[0] = jac1 * integrand2d    (xi1a, xi2a, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars, phip, xi, eta);
      else       res[0] = jac1 * integrand2d_SUB(xi1a, xi2a, xx1, xx2,       pr, a0*b0, muFrat, par, intVars, p.nf, a0);
      //res[0] = jac1 * integrand2d(t1, y1, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars);
    } else if(whichFunction == 2) {
      res[0] = jac2 * integrand2d(xi1b, xi2b, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars, phip, xi, eta);
      //res[0] = jac2 * integrand2d(t2, y2, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars);
    }
    //
    if(res[0]!=res[0]) {
      cout << "Result is nan.  Jacobian = " << jac1 << endl; // "  Integrand = " << integrand2d    (xi1a, xi2a, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars, phip, xi, eta) << endl;
      res[0]=0;  // tappullo!!
      /*
      cout << whichFunction << setw(12) << x[0] << setw(12) << x[1] << setw(12) << x[2];
      cout << setw(18) << t1 << setw(12) << y1 << setw(12) << t2 << setw(12) << y2 << setw(12) << rapY << setw(12) << xx1 << setw(12) << xx2 << setw(12) << xx;
      cout << endl;
      //*/
    } else {
      //cout << x[0] << "   " << x[1] << "   " << x[2] << "   " << res[0] << endl;
    }
    return 0;
  }
  int integrandCUBAx0(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    return integrandCUBAx(ndim, x, ncomp, res, pars, 0);
  }
  int integrandCUBAx1(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    return integrandCUBAx(ndim, x, ncomp, res, pars, 1);
  }
  int integrandCUBAx2(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    return integrandCUBAx(ndim, x, ncomp, res, pars, 2);
  }

  struct s_prdtvXX { s_prdtv pars; double *xx; int dim; };
  double test_integrandCUBAx1(double x, void *p) {
    s_prdtvXX intparsXX = *(s_prdtvXX*) p;
    int dim = intparsXX.dim;
    const int ncomp = 1;
    double cres[ncomp];
    double *xx = new double[dim];
    for(int i=0; i<dim; i++) xx[i] = intparsXX.xx[i];
    xx[dim-1] = x; // integrate the last variable: phip for SQ, phi for PairSUB1
    integrandCUBAx1(&dim, xx, &ncomp, cres, &intparsXX.pars);
    return cres[0];
  }
    double test_integrandCUBAx1_partiallyintegrated(double *xx, s_prdtv intpars, int dim) {
    s_prdtvXX intparsXX = { intpars, xx, dim };
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double prec = 1e-4;
    int status;
    double res = integration_gauss(test_integrandCUBAx1, 0, 1, prec, NULL, &intparsXX, &status);
    if(status) res = 0;
    gsl_set_error_handler (old_handler);
    return res;
  }


  double integral1d_DeltaFRes(s_prdtv intpars) {
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double prec = 1e-4;
    int status;
    double res = integration_gauss(integrandGSLx, 0, 1, prec, NULL, &intpars, &status);
    if(status) res = 0;
    gsl_set_error_handler (old_handler);
    return res;
  }
  double integralDd_DeltaFRes(s_prdtv intpars, int dim) {
    const int ncomp = 1;
    double cres[ncomp], error[ncomp], prob[ncomp];
    double epsrel=1e-2, epsabs=1.e-10;
    int last = 4;
    int verbose = 0;
    int nregions, neval, fail;
    //int mineval = 0, maxeval = 300000*pow(5,dim);
    int mineval = 0, maxeval = 10000*pow(5,dim);
    //int mineval = 5000, maxeval = 1000000;
    int nstart = 50000, nincrease = 20000, nbatch = 20000, gridno = 1;
    //int nstart = 20000, nincrease = 40000, nbatch = 1000, gridno = 5;
    //int nstart = 40000, nincrease = 40000, nbatch = 1000, gridno = 5;
    nregions=1; // to remove the warning
    int seed = 0;
    double result = 0;
    //int cores = 1, pmax= 10000;
    //cubacores(&cores,&pmax);
    //
    /*
    double N=1000;
    for(int j=0; j<dim; j++) {
      //if(j!=0) continue;
      cout << "--- integrand " << j << " ---" << endl;
      ostringstream ofilename;
      ofilename << "output/testintegrandPair" << j << ".dat";
      ofstream ofile(ofilename.str());
      for(int i=0; i<N; i++) {
	double x = i/(N-1.);
	double *xx = new double[dim];
	for(int k=0; k<dim; k++) xx[k] = 0.1*(0.05+k);
	//xx[dim-1] = 0.00002;
	//for(int k=0; k<dim; k++) xx[k] = 0.15*(0.05+k);
	//xx[dim-1] = 0.0002;
	xx[j] = x;
	//cout << setw(4) << j << setw(8) << i << endl;
	integrandCUBAx1(&dim, xx, &ncomp, cres, &intpars);
	//if(i%10==0) cout << cres[0] << endl;
	//cres[0] = test_integrandCUBAx1_partiallyintegrated(xx, intpars, dim);
	ofile << setw(12) << x << setw(16) << cres[0] << endl;
      }
    }
    exit(1);
    //*/
    /*
    double N=10000;
    for(int j=0; j<dim; j++) {
      //if(j!=1) continue;
      cout << "--- integrand " << j << " ---" << endl;
      ostringstream ofilename;
      ofilename << "output/testintegrand" << j << ".dat";
      ofstream ofile(ofilename.str());
      for(int i=0; i<N; i++) {
	double x = i/(N-1.);
	double *xx = new double[dim];
	for(int k=0; k<dim; k++) xx[k] = 0.1*(0.05+k);
	xx[dim-1] = 0.00002;
	//for(int k=0; k<dim; k++) xx[k] = 0.15*(0.05+k);
	//xx[dim-1] = 0.0002;
	xx[j] = x;
	//cout << setw(4) << j << setw(8) << i << endl;
	integrandCUBAx1(&dim, xx, &ncomp, cres, &intpars);
	//if(i%10==0) cout << cres[0] << endl;
	//cres[0] = test_integrandCUBAx1_partiallyintegrated(xx, intpars);
	ofile << setw(12) << x << setw(16) << cres[0] << endl;
      }
    }
    exit(1);
    //*/
    //
    //Vegas(dim, ncomp, integrandCUBAx0, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, nstart, nincrease, nbatch, gridno, NULL, NULL, &neval, &fail, cres, error, prob);
    //Suave(dim, ncomp, integrandCUBAx0, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 1, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
    //Cuhre(dim, ncomp, integrandCUBAx0, &intpars, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
    //result = cres[0];
    //
    //Cuhre(dim, ncomp, integrandCUBAx1, &intpars, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
    Vegas(dim, ncomp, integrandCUBAx1, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, nstart, nincrease, nbatch, gridno, NULL, NULL, &neval, &fail, cres, error, prob);
    //Suave(dim, ncomp, integrandCUBAx1, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 1, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
    result += cres[0];
    //
    if(intpars.p != QQbarPairM2Yq && intpars.p != QQbarPairM2Yq_SUB1 && intpars.p != QQbarPairM2Yq_SUB4 && intpars.p != QQbarPairM2Yq_SUBc && intpars.p != QQbarPairM2Yq_v2 && intpars.p != QQbarPairM2Y_v2 && intpars.p != QQbarPairM2Y && intpars.p != QQbarPairM2YAux && intpars.p != QQbarPairM2YLO && intpars.p != QQbarSingleQuarkYq && intpars.p != QQbarSingleQuarkYq_v2) {
      cout << "ciccio" << endl;
      //Cuhre(dim, ncomp, integrandCUBAx2, &intpars, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
      Vegas(dim, ncomp, integrandCUBAx2, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, nstart, nincrease, nbatch, gridno, NULL, NULL, &neval, &fail, cres, error, prob);
      //Suave(dim, ncomp, integrandCUBAx2, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 1, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
      result += cres[0];
    }
    //
    if(fail == -1)  cout << "  ERROR: dimension out of range." << endl;
    else if(fail>0) cout << "  ERROR: precision not reached." << endl;
    cout << result << " +/- " << error[0] << "   dimension: " << dim << endl;
    return result;
  }




    double newDeltaFRes(double x, double as, int nf, xTableU *tU, process proc, int dim, bool sub=false, double muFrat=1, void *pars=NULL, double rapY=0) {
    double b0 = beta0(nf);
    //
    s_prdtv intpars = { proc, x, as, b0, muFrat, rapY, tU, pars, nf, sub };
    double res = 0;
    if(dim==1)
      res = integral1d_DeltaFRes(intpars);
    else
      res = integralDd_DeltaFRes(intpars, dim);
    //
    return res;
  }






  // QQbar pair, triple differential
  double DeltaQQbarPairM2YqtRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    QQbarPair *QQbar = (QQbarPair*) pars;
    double rapY = QQbar->GetRapidity();
    //double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_v2, 4, 1, muFrat, pars, rapY);
    double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq, 4, false, muFrat, pars, rapY);
    //res -= newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUB1, 4, true, muFrat, pars, rapY);
    //res -= newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUB4, 4, true, muFrat, pars, rapY);
    //res -= newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUBc, 2, true, muFrat, pars, rapY);
    return res;
  }
  double DeltaQQbarPairM2YqtSUB(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    QQbarPair *QQbar = (QQbarPair*) pars;
    double rapY = QQbar->GetRapidity();
    double res = 0;
    res -= newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUB1, 4, true, muFrat, pars, rapY);
    res -= newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUB4, 4, true, muFrat, pars, rapY);
    res -= newDeltaFRes(x, as, nf, tU, QQbarPairM2Yq_SUBc, 2, true, muFrat, pars, rapY);
    return -res;
  }
  double DeltaQQbarPairM2YqtResAux(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    QQbarPair *QQbar = (QQbarPair*) pars;
    double rapY = QQbar->GetRapidity();
    double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2YqAux, 1, false, muFrat, pars, rapY);  /// fix dimension!
    return res;
  }

  // QQbar pair, double differential
  double DeltaQQbarPairM2YRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    QQbarPair *QQbar = (QQbarPair*) pars;
    double rapY = QQbar->GetRapidity();
    double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2Y, 5, false, muFrat, pars, rapY);
    //double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2Y_v2, 5, false, muFrat, pars, rapY);
    return res;
  }
  double DeltaQQbarPairM2YResAux(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    QQbarPair *QQbar = (QQbarPair*) pars;
    double rapY = QQbar->GetRapidity();
    double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2YAux, 2, false, muFrat, pars, rapY);
    return res;
  }
  double DeltaQQbarPairM2YLO(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    QQbarPair *QQbar = (QQbarPair*) pars;
    double rapY = QQbar->GetRapidity();
    double res = newDeltaFRes(x, as, nf, tU, QQbarPairM2YLO, 3, false, muFrat, pars, rapY);
    return res;
  }






  // QQbar single quark, double Y-pt differential
  double DeltaQQbarSingleQuarkYqtRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    QQbarSingleQuark *QQbar = (QQbarSingleQuark*) pars;
    double rapY = QQbar->GetRapidity();
    //double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYq, 5, false, muFrat, pars, rapY);
    double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYq_v2, 5, false, muFrat, pars, rapY);
    return res;
  }
  double DeltaQQbarSingleQuarkYqtResAux(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    QQbarSingleQuark *QQbar = (QQbarSingleQuark*) pars;
    double rapY = QQbar->GetRapidity();
    double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYqAux, 1, false, muFrat, pars, rapY);  /// fix dimension!
    return res;
  }

  // QQbar single quark, Y differential
  double DeltaQQbarSingleQuarkYRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    QQbarSingleQuark *QQbar = (QQbarSingleQuark*) pars;
    double rapY = QQbar->GetRapidity();
    double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkY, 1, false, muFrat, pars, rapY);  /// fix dimension!
    return res;
  }
  double DeltaQQbarSingleQuarkYResAux(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    QQbarSingleQuark *QQbar = (QQbarSingleQuark*) pars;
    double rapY = QQbar->GetRapidity();
    double res = newDeltaFRes(x, as, nf, tU, QQbarSingleQuarkYAux, 1, false, muFrat, pars, rapY);  /// fix dimension!
    return res;
  }


};
};

