#include "xTableU.hh"
#include "version.hh"


using namespace std;



namespace HELL {

  xTableU::xTableU(string filename) {
    ifstream infile(filename.c_str());
    if(!infile.good()) {
      cout << "\033[0;31m" << "HELL: Error reading table " << filename << "\033[0m" << endl;
      exit(0);
    }
    string vers;
    infile >> vers >> Np1 >> Np2 >> x_min >> x_mid >> x_max >> Nxi;
    if(vers!=VERSION) {
      std::cout << "\033[0;31m" << "HELL: Error! The U tables you are trying to read are version " << vers << " which is not compatible with the code version " << VERSION << "\033[0m" << std::endl;
      abort();
    }
    //Nxi--; // this patches a bad choice in the produced files  //// FIX THIS
    xx    = new double[Np1+Np2];
    xdUqg = new double[Np1+Np2];    
    for(int i=0; i<Np1+Np2; i++) {
      if(i<Np1) xx[i] = x_min * exp(i/(Np1-1.)*log(x_mid/x_min));
      else      xx[i] = x_mid + (i-Np1+1)*(x_max-x_mid)/(Np2-0.);
      infile >> xdUqg[i];
    }
    //}
    //void xTableU::Init() {
    xi_    = new double [Nxi];
    logxi_ = new double [Nxi];
    xdU    = new double*[Nxi];
    //xdDU   = new double*[Nxi];
    for(int m=0; m<Nxi; m++) {
      /*
      xdU[Nxi-1-m] = new double[Np1+Np2];
      infile >> xi_[Nxi-1-m];
      for(int i=0; i<Np1+Np2; i++) {
	infile >> xdU[Nxi-1-m][i];
      }
      //cout << Nxi-1-m << "   "<< xi_[Nxi-1-m] << "  " << xdU[Nxi-1-m][0] << "  " << xdU[Nxi-1-m][Np1+Np2-1] << endl;
      */
      xdU[m]  = new double[Np1+Np2];
      infile >> xi_[m];
      logxi_[m] = log(xi_[m]);
      for(int i=0; i<Np1+Np2; i++) {
	infile >> xdU[m][i];
      }
      //cout << Nxi-1-m << "   "<< xi_[Nxi-1-m] << "  " << xdU[Nxi-1-m][0] << "  " << xdU[Nxi-1-m][Np1+Np2-1] << endl;
    }
    /*
    for(int m=0; m<Nxi; m++) {
      xdDU[m] = new double[Np1+Np2];
      double tmpxi;
      infile >> tmpxi;  // should be xi_[m];
      for(int i=0; i<Np1+Np2; i++) {
	infile >> xdDU[m][i];
      }
      //cout << Nxi-1-m << "   "<< xi_[Nxi-1-m] << "  " << xdU[Nxi-1-m][0] << "  " << xdU[Nxi-1-m][Np1+Np2-1] << endl;
    }
    */
    infile.close();
    dUspline = new gsl_spline;
    dUspline = gsl_spline_alloc (gsl_interp_cspline, Nxi);
    //dUspline = gsl_spline_alloc (gsl_interp_linear, Nxi);
  }
  double xTableU::interpolate(double x) {
    if(x>1 || x<0) {
      cout << "\033[0;31m" << "HELL: Error! Requesting evolution function U for unphysical value of x=" << x << " outside the physical range 0<x<=1" << "\033[0m" << endl;
      exit(45);
    }
    if(x>x_max) {
      cout << "\033[0;31m" << "HELL: Warning! Extrapolating U out of interpolation range: x=" << x << " > x_max=" << x_max << "\033[0m" << endl;
      x = x_max;
    }
    if(x<x_min) {
      cout << "\033[0;31m" << "HELL: Warning! Extrapolating U out of interpolation range: x=" << x << " < x_min=" << x_min << "\033[0m" << endl;
      x = x_min;
    }
    double ii;
    if(x<x_mid) ii = (Np1-1.)*log(x/x_min)/log(x_mid/x_min);
    else        ii = Np1-1.+Np2*(x-x_mid)/(x_max-x_mid);
    return ii;
  }
  double xTableU::evalU(double x, double xi, double &deriv) {
    double ii = interpolate(x);
    int i = int(ii);
    if(i<0 || ii<i) {
      cout << "\033[0;31m" << "HELL: Error! This should never happen" << "\033[0m" << endl;
      abort();
    }
    double *xidU;
    xidU = new double[Nxi];
    for(int m=0; m<Nxi; m++) {
      xidU[m] = xdU[m][i] + (i==Np1+Np2-1 ? 0 : (ii-i)*(xdU[m][i+1]-xdU[m][i]));
    }
    // spline setup
    double dU = 0;
    deriv = 0;
    if(xi<xi_[0]) dU = xidU[0];
    else if(xi> xi_[Nxi-1]) {
      cout << "Warning: extrapolating U at large xi" << endl;
      dU = xidU[Nxi-1];
      //deriv=??
    }
    else {
      gsl_spline_init (dUspline, logxi_, xidU, Nxi);
      gsl_interp_accel *acc = gsl_interp_accel_alloc ();
      dU    = gsl_spline_eval      (dUspline, log(xi), acc);
      deriv = gsl_spline_eval_deriv(dUspline, log(xi), acc) /xi;
      gsl_interp_accel_free (acc);
    }
    delete[] xidU;
    deriv /= x;
    dU    /= x;
    deriv *= pow(1-x,2);  // damping
    dU    *= pow(1-x,2);  // damping
    return dU;
  }
  /*
  // this is not better than computing the derivative numerically, because the U' oscillates more than U, so the interpolation is better for U than U'
  double xTableU::evalDU(double x, double xi) {
    double ii = interpolate(x);
    int i = int(ii);
    if(i<0 || ii<i) {
      cout << "\033[0;31m" << "HELL: Error! This should never happen" << "\033[0m" << endl;
      abort();
    }
    double *xidDU;
    xidDU = new double[Nxi];
    for(int m=0; m<Nxi; m++) {
      xidDU[m] = xdDU[m][i] + (i==Np1+Np2-1 ? 0 : (ii-i)*(xdDU[m][i+1]-xdDU[m][i]));
    }
    // spline setup
    double dDU = 0;
    if(xi<=xi_[0]) dDU = xidDU[0];
    else if(xi> xi_[Nxi-1]) {
      cout << "Warning: extrapolating U at large xi" << endl;
      dDU = xidDU[Nxi-1];
      //deriv=??
    }
    else {
      gsl_spline_init (dUspline, logxi_, xidDU, Nxi);
      gsl_interp_accel *acc = gsl_interp_accel_alloc ();
      dDU    = gsl_spline_eval      (dUspline, log(xi), acc);
      gsl_interp_accel_free (acc);
    }
    delete[] xidDU;
    dDU    /= x;
    dDU    *= pow(1-x,2);  // damping
    return dDU;
  }
  */
  double xTableU::evalUqg(double x) {
    double ii = interpolate(x);
    int i = int(ii);
    if(i<0 || ii<i) {
      cout << "\033[0;31m" << "HELL: Error! This should never happen" << "\033[0m" << endl;
      abort();
    }
    double dU;
    dU = xdUqg[i] + (i==Np1+Np2-1 ? 0 : (ii-i)*(xdUqg[i+1]-xdUqg[i]));
    dU /= x;
    dU *= pow(1-x,2);  // damping
    return dU;
  }



};

